#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#Created on Wed Mar  8 21:51:38 2023
#@author: Ajit Johnson Nirmal
#Run CSPOT Pipeline

"""
!!! abstract "Short Description"
    The csPipeline function is simply a wrapper for the following functions:  
    - csPredict  
    - generateCSScore  
    - csObject  
    - cspot  
      
    Typically, in production settings, `csPipeline` would be utilized, whereas 
    step-by-step analysis would be employed for troubleshooting, model validation, 
    and similar tasks that necessitate greater granularity or control.
      
    Please refer to the individual function documentation for parameter tuning.

## Function
"""

# libs
import inspect
import argparse
import pathlib
import os

# CSPOT Functions
if __name__ == '__main__':
    from csPredict import csPredict
    from generateCSScore import generateCSScore
    from csObject import csObject
    from cspot import cspot
else:
    from .csPredict import csPredict
    from .generateCSScore import generateCSScore
    from .csObject import csObject
    from .cspot import cspot



# Function
def csPipeline (**kwargs):   
    """
Parameters:
    imagePath (str):  
        The path to the .tif file that needs to be processed. 
     
    csModelPath (str):  
        The path to the `cspotModel` folder. 

    markerChannelMapPath (str, optional):  
        The path to the marker panel list, which contains information about the markers used in the image. This argument is required.

    segmentationMaskPath (str):
        Supply the path of the pre-computed segmentation mask.

    spatialTablePath (list):
        Provide a list of paths to the single-cell spatial feature tables, ensuring each image has a unique path specified.
         
    projectDir (str):  
        The path to the output directory where the processed images (`probabilityMasks`) will be saved.

    verbose (bool, optional):
        If True, print detailed information about the process to the console.  
        
    markerColumnName (str, optional):  
        The name of the column in the marker panel list that contains the marker names. The default value is 'marker'.
     
    channelColumnName (str, optional):  
        The name of the column in the marker panel list that contains the channel names. The default value is 'channel'.
     
    modelColumnName (str, optional):  
        The name of the column in the marker panel list that contains the model names. The default value is 'cspotmodel'.

    GPU (int, optional):  
        An optional argument to explicitly select the GPU to use. The default value is -1, meaning that the GPU will be selected automatically.

    feature (str, optional):
        Calculates the `mean` or `median` CSPOT Score for each cell.

    markerNames (list, optional):
        The program searches for marker names in the meta data (description section)
        of the tiff files created by `csPredict` by default. If the meta data
        is lost due to user modifications, provide the marker names for each
        channel/layer in the `probabilityMaskPath` here.

    CellId (str, optional):
        Specify the column name that holds the cell ID (a unique name given to each cell).

    uniqueCellId (bool, optional):
        The function generates a unique name for each cell by combining the CellId and imageid.
        If you don't want this, pass False. In such case the function will default to using just the CellId.
        However, make sure CellId is unique especially when loading multiple images together.

    split (string, optional):
        The spatial feature table generally includes single cell expression data
        and meta data such as X, Y coordinates, and cell shape size. The CSPOT
        object separates them. Ensure that the expression data columns come first,
        followed by meta data columns. Provide the column name that marks the split,
        i.e the column name immediately following the expression data.

    removeDNA (bool, optional):
        Exclude DNA channels from the final output. The function searches for
        column names containing the string `dna` or `dapi`. 

    remove_string_from_name (string, optional):
        Cleans up channel names by removing user specified string from all marker
        names.

    csScore (str, optional):
        Include the label used for saving the `csScore` within the CSPOT object.

    minAbundance (float, optional):
        Specify the minimum percentage of cells that should express a specific
        marker in order to determine if the marker is considered a failure.
        A good approach is to consider the lowest percentage of rare cells
        expected within the dataset.

    percentiles (list, optional):
        Specify the interval of percentile levels of the expression utilized to intialize
        the GMM. The cells falling within these percentiles are utilized to distinguish
        between negative cells (first two values) and positive cells (last two values).

    dropMarkers (list, optional):
        Specify a list of markers to be removed from the analysis, for
        example: `["background_channel1", "background_channel2"]`. 

    RobustScale (bool, optional):
        When set to True, the data will be subject to Robust Scaling before the
        Gradient Boosting Classifier is trained. 

    log (bool, optional):
        Apply `log1p` transformation on the data, unless it has already been log
        transformed in which case set it to `False`. 

    stringentThreshold (bool, optional):
        The Gaussian Mixture Model (GMM) is utilized to distinguish positive and 
        negative cells by utilizing csScores. The stringentThreshold can be utilized 
        to further refine the classification of positive and negative cells. 
        By setting it to True, cells with csScore below the mean of the negative 
        distribution and above the mean of the positive distribution will be 
        labeled as true negative and positive, respectively.
        
    x_coordinate (str, optional):
        The column name in `single-cell spatial table` that records the
        X coordinates for each cell. 

    y_coordinate (str, optional):
        The column name in `single-cell spatial table` that records the
        Y coordinates for each cell.

    imageid (str, optional):
        The name of the column that holds the unique image ID. 

    random_state (int, optional):
        Seed used by the random number generator. 

    rescaleMethod (string, optional):
        Choose between `sigmoid` and `minmax`.

    label (str, optional):
        Assign a label for the object within `adata.uns` where the predictions
        from CSPOT will be stored. 
    
    
Returns:
    csObject (anndata):
        Returns a csObject with predictions of all positve and negative cells. 

Example:
        ```python
        
        # Path to all the files that are necessary files for running the 
        CSPOT Prediction Algorithm (broken down based on sub functions)
        projectDir = '/Users/aj/Documents/cspotExampleData'

        # gatorPredict related paths
        imagePath = projectDir + '/image/exampleImage.tif'
        markerChannelMapPath = projectDir + '/markers.csv'
        csModelPath = projectDir + '/manuscriptModels/'
        
        # Generate generateGatorScore related paths
        segmentationPath = projectDir + '/segmentation/exampleSegmentationMask.tif'
        
        # gatorObject related paths
        spatialTablePath = projectDir + '/quantification/exampleSpatialTable.csv'
        
        # Run the pipeline
        cs.csPipeline(   
                    # parameters for gatorPredict function
                    imagePath=imagePath,
                    csModelPath=csModelPath,
                    markerChannelMapPath=markerChannelMapPath,

                    # parameters for generateGatorScore function
                    segmentationMaskPath=segmentationPath,

                    # parameters for gatorObject function
                    spatialTablePath=spatialTablePath,

                    # parameters to run gator function
                    # ..

                    # common parameters
                    verbose=False,
                    projectDir=projectDir)
        
        # Same function if the user wants to run it via Command Line Interface
        python csPipeline.py \
                --imagePath /Users/aj/Documents/cspotExampleData/image/exampleImage.tif \
                --csModelPath /Users/aj/Documents/cspotExampleData/CSPOT/cspotModel/ \
                --markerChannelMapPath /Users/aj/Documents/cspotExampleData/markers.csv \
                --segmentationMaskPath /Users/aj/Documents/cspotExampleData/segmentation/exampleSegmentationMask.tif \
                --spatialTablePath /Users/aj/Documents/cspotExampleData/quantification/exampleSpatialTable.csv \
                --projectDir /Users/aj/Documents/cspotExampleData
        ```
                
         
    """
    
    ##########################################################################
    # STEP: 1 :- PREDICT
    ##########################################################################
    function1_args = inspect.signature(csPredict).parameters.keys()
    # Extract only the arguments that csPredict expects from the keyword arguments
    function1_kwargs = {k: kwargs[k] for k in kwargs if k in function1_args}
    # Call csPredict with the extracted arguments
    csPredict (**function1_kwargs)
    
    
    ##########################################################################
    # STEP: 2 :- generateCSScore
    ##########################################################################
    
    # derive the probability mask path
    probPath = pathlib.Path(kwargs['projectDir'] + '/CSPOT/csPredict/')
    fileName = os.path.basename(kwargs['imagePath'])
    fileNamePrefix = fileName.split(os.extsep, 1)
    probabilityMaskPath = str(probPath / (fileNamePrefix[0] + '_cspotPredict.ome.tif'))
    
    # extract key words for generateCSScore
    function2_args = inspect.signature(generateCSScore).parameters.keys()
    function2_kwargs = {k: kwargs[k] for k in kwargs if k in function2_args}
    generateCSScore (probabilityMaskPath=probabilityMaskPath, **function2_kwargs)


    ##########################################################################
    # STEP: 3 :- Generate csObject
    ##########################################################################
     
    # derive the path to CSPOT scores
    gPath = pathlib.Path(kwargs['projectDir'] + '/CSPOT/csScore/')
    file_name = pathlib.Path(probabilityMaskPath).stem + '.csv'
    csScorePath = str(gPath / file_name)
    
    # extract key words for csObject
    function3_args = inspect.signature(csObject).parameters.keys()
    function3_kwargs = {k: kwargs[k] for k in kwargs if k in function3_args}
    csObject (csScorePath=csScorePath, **function3_kwargs)
    
    ##########################################################################
    # STEP: 4 :- Run CSPOT Algorithm
    ##########################################################################
    
    # derive the path to CSPOT object
    oPath = pathlib.Path(kwargs['projectDir'] + '/CSPOT/csObject/')
    file_name = pathlib.Path(csScorePath).stem + '.h5ad'
    csObjectPath = str(oPath / file_name)
    
    # extract key words for running CSPOT
    function4_args = inspect.signature(cspot).parameters.keys()
    function4_kwargs = {k: kwargs[k] for k in kwargs if k in function4_args}    
    cspot (csObject=csObjectPath, **function4_kwargs)



if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Run csPipeline function')
    parser.add_argument('--imagePath', type=str, help='path to the .tif file that needs to be processed')
    parser.add_argument('--csModelPath', type=str, help='path to the `cspotModel` folder')
    parser.add_argument('--markerChannelMapPath', type=str, help='path to the marker panel list')
    parser.add_argument('--segmentationMaskPath', type=str, help='path to the pre-computed segmentation mask')
    parser.add_argument('--spatialTablePath', type=str, nargs='+', help='list of paths to the single-cell spatial feature tables')
    parser.add_argument('--projectDir', type=str, help='path to the output directory where the processed images (`probabilityMasks`) will be saved')
    parser.add_argument('--verbose', type=bool, default=True, help='print detailed information about the process to the console')
    
    parser.add_argument('--markerColumnName', type=str, default='marker', help='name of the column in the marker panel list that contains the marker names')
    parser.add_argument('--channelColumnName', type=str, default='channel', help='name of the column in the marker panel list that contains the channel names')
    parser.add_argument('--modelColumnName', type=str, default='cspotmodel', help='name of the column in the marker panel list that contains the model names')
    parser.add_argument('--GPU', type=int, default=-1, help='explicitly select the GPU to use')
    
    parser.add_argument('--feature', type=str, choices=['mean', 'median'], help='calculates the `mean` or `median` CSPOT Score for each cell')
    parser.add_argument('--markerNames', type=str, nargs='+', help='provide the marker names for each channel/layer in the `probabilityMaskPath`')
    
    parser.add_argument('--CellId', type=str, default='CellID', help='Specify the column name that holds the cell ID.')
    parser.add_argument('--uniqueCellId', type=bool, default=False, help='the function generates a unique name for each cell by combining the CellId and imageid')
    parser.add_argument('--split', type=str, default='X_centroid', help='Provide the column name that marks the split between expression data and meta data.')
    parser.add_argument('--removeDNA', type=bool, default=True, help='exclude DNA channels from the final output')
    parser.add_argument('--remove_string_from_name', type=str, help='cleans up channel names by removing user specified string from all marker names')
    parser.add_argument('--dropMarkers', type=list, default=None, help='Specify a list of markers to be removed from the analysis')
    parser.add_argument('--csScore', type=str, default='csScore', help='Include the label used for saving the `csScore` within the CSPOT object')
    parser.add_argument('--minAbundance', type=float, default=0.002, help='Specify the minimum percentage of cells that should express a specific marker in order to determine if the marker is considered a failure')
    parser.add_argument('--percentiles', type=list, default=[1, 20, 80, 99], help='Specify the interval of percentile levels of the expression utilized to intialize the GMM')
    parser.add_argument('--RobustScale', type=bool, default=False, help='When set to True, the data will be subject to Robust Scaling before the Gradient Boosting Classifier is trained')
    parser.add_argument('--log', type=bool, default=True, help='Apply `log1p` transformation on the data, unless it has already been log transformed in which case set it to `False`')
    parser.add_argument('--stringentThreshold', type=bool, default=False, help='Threshold to refine the classification of positive and negative cells')
    parser.add_argument('--x_coordinate', type=str, default='X_centroid', help='The column name in `single-cell spatial table` that records the X coordinates for each cell')
    parser.add_argument('--y_coordinate', type=str, default='Y_centroid', help='The column name in `single-cell spatial table` that records the Y coordinates for each cell')
    parser.add_argument('--imageid', type=str, default='imageid', help='The name of the column that holds the unique image ID')
    parser.add_argument('--random_state', type=int, default=0, help='Seed used by the random number generator')
    parser.add_argument('--rescaleMethod', type=str, default='minmax', help='Choose between `sigmoid` and `minmax`')
    parser.add_argument('--label', type=str, default='cspotOutput', help='Assign a label for the object within `adata.uns` where the predictions from CSPOT will be stored')

    
    args = parser.parse_args()
    csPipeline(imagePath=args.imagePath, 
                  csModelPath=args.csModelPath, 
                  markerChannelMapPath=args.markerChannelMapPath, 
                  segmentationMaskPath=args.segmentationMaskPath, 
                  spatialTablePath=args.spatialTablePath, 
                  projectDir=args.projectDir, 
                  verbose=args.verbose,
                  markerColumnName=args.markerColumnName, 
                  channelColumnName=args.channelColumnName, 
                  modelColumnName=args.modelColumnName, 
                  GPU=args.GPU, 
                  feature=args.feature, 
                  markerNames=args.markerNames, 
                  CellId=args.CellId, 
                  uniqueCellId=args.uniqueCellId, 
                  split=args.split, 
                  removeDNA=args.removeDNA, 
                  remove_string_from_name=args.remove_string_from_name, 
                  dropMarkers=args.dropMarkers,
                  csScore=args.csScore, 
                  minAbundance=args.minAbundance, 
                  percentiles=args.percentiles,
                  RobustScale=args.RobustScale, 
                  log=args.log, 
                  stringentThreshold=args.stringentThreshold, 
                  x_coordinate=args.x_coordinate, 
                  y_coordinate=args.y_coordinate, 
                  imageid=args.imageid, 
                  random_state=args.random_state, 
                  rescaleMethod=args.rescaleMethod, 
                  label=args.label)





import subprocess, re
import numpy as np

def pick_gpu_lowest_memory():
    output = subprocess.Popen("nvidia-smi", stdout=subprocess.PIPE, shell=True).communicate()[0]
    output=output.decode("ascii")
    gpu_output = output[output.find("Memory-Usage"):]
        # lines of the form
        # |    0      8734    C   python                                       11705MiB |
    memory_regex = re.compile(r"[|]\s+?\D+?.+[ ](?P<gpu_memory>\d+)MiB /")
    rows = gpu_output.split("\n")
    result=[]
    for row in gpu_output.split("\n"):
        m = memory_regex.search(row)
        if not m:
            continue
        gpu_memory = int(m.group("gpu_memory"))
        result.append(gpu_memory)
    return np.argsort(result)[0]


from os.path import *
import pickle


def pathjoin(p,ne): # '/path/to/folder', 'name.extension' (or a subfolder)
    return join(p,ne)

def saveData(data,path):
    print('saving data')
    dataFile = open(path, 'wb')
    pickle.dump(data, dataFile)

def loadData(path):
    print('loading data')
    dataFile = open(path, 'rb')
    return pickle.load(dataFile)


# -*- coding: utf-8 -*-
#Created on Thu Sep  1 16:54:39 2022
#@author: Ajit Johnson Nirmal
#Function to calculate the mean/median probability

"""
!!! abstract "Short Description"
    The `generateCSScore` function calculates `CSPOT Score` for each cell by using 
    both the generated probability masks and pre-computed segmentation masks as inputs


## Function
"""

# Libs
import tifffile
import pandas as pd
from skimage import measure
#import ast
import xml.etree.ElementTree as ET
import numpy as np
import pathlib
import os
import argparse

# Function
def generateCSScore (probabilityMaskPath,
                         segmentationMaskPath,
                         feature='median',
                         verbose=True,
                         markerNames=None,
                         projectDir=None):

    """
Parameters:
    probabilityMaskPath (str):
        Supply the path of the probability map image produced by `dlModelPredict`.

    segmentationMaskPath (str):
        Supply the path of the pre-computed segmentation mask.

    feature (str, optional):
        Calculates the `mean` or `median` CSPOT Score for each cell.

    verbose (bool, optional):
        If True, print detailed information about the process to the console.  

    markerNames (list, optional):
        The program searches for marker names in the meta data (description section)
        of the tiff files created by `csPredict` by default. If the meta data
        is lost due to user modifications, provide the marker names for each
        channel/layer in the `probabilityMaskPath` here.

    projectDir (str, optional):
        Provide the path to the output directory. The result will be located at
        `projectDir/CSPOT/csScore/`.

Returns:
    CSV (dataframe):
        The `.csv` file containing the `csScore` is stored in the provided projectDir.

Example:
        ```python
        
        # global path
        projectDir = '/Users/aj/Documents/cspotExampleData'
        
        # Path to all the files that are necessary files for running generateCSScore
        segmentationPath = projectDir + '/segmentation/exampleSegmentationMask.tif'
        probabilityMaskPath = projectDir + '/CSPOT/csPredict/exampleImage_cspotPredict.ome.tif'
        
        cs.generateCSScore(probabilityMaskPath=probabilityMaskPath,
                      segmentationMaskPath=segmentationPath,
                      feature='median',
                      projectDir=projectDir)
        
        # Same function if the user wants to run it via Command Line Interface
        python generateCSScore.py \
            --probabilityMaskPath /Users/aj/Documents/cspotExampleData/CSPOT/csPredict/exampleImage_cspotPredict.ome.tif \
            --segmentationMaskPath /Users/aj/Documents/cspotExampleData/segmentation/exampleSegmentationMask.tif \
            --projectDir /Users/aj/Documents/cspotExampleData
        
        ```

    """

    # read the seg mask
    segM = tifffile.imread(pathlib.Path(segmentationMaskPath))
    probM = tifffile.imread(pathlib.Path(probabilityMaskPath))

    #probs = []
    #for i in range(len(probM)):
    #    pospix = len(probM[i][(probM[i] / 255) > 0.5]) / (probM[i].shape[0] * probM[i].shape[1])
    #    probs.append(pospix)

    if len(probM.shape) > 2:
        probM = np.moveaxis(probM, 0, -1)

    def median_intensity(mask, img):
        return np.median(img[mask])

    # quantify
    if verbose is True:
        print("Quantifying the probability masks")
    quantTable = pd.DataFrame(measure.regionprops_table(segM, intensity_image=probM,
                                                        properties=['label','mean_intensity'],
                                                        extra_properties=[median_intensity])).set_index('label')

    # keep only median
    if feature == 'median':
        quantTable = quantTable.filter(regex='median')
    if feature == 'mean':
        quantTable = quantTable.filter(regex='mean')

    # read the channel names from the tiffile
    tiff = tifffile.TiffFile(probabilityMaskPath)
    try:
        root = ET.fromstring(tiff.pages[0].description)
        # parse the ome XML
        namespace = None
        for elem in root.iter():
            if "Channel" in elem.tag:
                namespace = {"ome": elem.tag.split("}")[0][1:]}
                break
        channel_names = [channel.get("Name") for channel in root.findall(".//ome:Channel", namespace)]
        quantTable.columns = channel_names
        #omexml_string = ast.literal_eval(tiff.pages[0].description)
        #channel_names = omexml_string['Channel']['Name']
    except:
        pass
    if markerNames is not None:
        channel_names = markerNames
    else:
        channel_names = list(quantTable.columns)

    # assign channel names
    quantTable.columns = channel_names

    # build a division vector
    # this is to make sure the low probs are not amplified; chosing 154 as it is 0.6P
    #div_val = []
    #for i in quantTable.columns:
    #    div_val.append(255 if quantTable[i].max() < 154 else  quantTable[i].max())

    # conver to prob
    #quantTable = quantTable / div_val
    quantTable = quantTable / 255

    # identify markers that failed
    #sns.distplot(quantTable['ECAD'])


    # if projectDir is given
    if projectDir is None:
        projectDir = os.getcwd()

    # final path to save results
    finalPath = pathlib.Path(projectDir + '/CSPOT/csScore/')
    if not os.path.exists(finalPath):
        os.makedirs(finalPath)

    # file name
    file_name = pathlib.Path(probabilityMaskPath).stem + '.csv'
    quantTable.to_csv(finalPath / file_name)
    
    # Finish Job
    if verbose is True:
        print('csScore is ready, head over to' + str(projectDir) + '/CSPOT/csScore" to view results')


# Make the Function CLI compatable
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Calculate csScore for a probability map and segmentation mask.')
    parser.add_argument('--probabilityMaskPath', type=str, help='Path of the probability map image produced by dlModelPredict.')
    parser.add_argument('--segmentationMaskPath', type=str, help='Path of the pre-computed segmentation mask.')
    parser.add_argument('--feature', type=str, default='median', help='Calculates the mean or median csScore for each cell.')
    parser.add_argument("--verbose", type=bool, default=True, help="If True, print detailed information about the process to the console.")       
    parser.add_argument('--markerNames', nargs='+', help='List of marker names for each channel/layer in the probabilityMaskPath.')
    parser.add_argument('--projectDir', type=str, help='Path to the output directory.')
    args = parser.parse_args()
    generateCSScore(probabilityMaskPath=args.probabilityMaskPath,
                    segmentationMaskPath=args.segmentationMaskPath,
                    feature=args.feature,
                    verbose=args.verbose,
                    markerNames=args.markerNames,
                    projectDir=args.projectDir)



# -*- coding: utf-8 -*-
#Created on Wed Jan 25 15:53:02 2023
#@author: ajn16
#CSPOT Phenotype


"""
!!! abstract "Short Description"
    The csPhenotype function requires a phenotype workflow document to guide 
    the algorithm in performing classification.  
    
    The phenotype workflow document is imported as a `dataframe` and passed to the 
    `phenotype` argument. It should follow the following structure:  
        
    (1) The `first column` has to contain the cell that are to be classified.  
    (2) The `second column` indicates the phenotype a particular cell will be assigned 
    if it satifies the conditions in the row.  
    (3) `Column three` and onward represent protein markers. If the protein marker 
    is known to be expressed for that cell type, then it is denoted by either `pos`, 
    `allpos`. If the protein marker is known to not express for a cell type it 
    can be denoted by `neg`, `allneg`. If the protein marker is irrelevant or 
    uncertain to express for a cell type, then it is left empty. `anypos` and 
    `anyneg` are options for using a set of markers and if any of the marker is 
    positive or negative, the cell type is denoted accordingly.  
        
    To give users maximum flexibility in identifying desired cell types, 
    we have implemented various classification arguments as described above 
    for strategical classification. They include  
    
    - allpos
    - allneg
    - anypos
    - anyneg
    - pos
    - neg
    
    `pos` : "Pos" looks for cells positive for a given marker. If multiple 
    markers are annotated as `pos`, all must be positive to denote the cell type. 
    For example, a Regulatory T cell can be defined as `CD3+CD4+FOXP3+` by passing 
    `pos` to each marker. If one or more markers don't meet the criteria (e.g. CD4-), 
    the program will classify it as `Likely-Regulatory-T cell`, pending user 
    confirmation. This is useful in cases of technical artifacts or when cell 
    types (such as cancer cells) are defined by marker loss (e.g. T-cell Lymphomas).  
    
    `neg` : Same as `pos` but looks for negativity of the defined markers.  
    
    `allpos` : "Allpos" requires all defined markers to be positive. Unlike 
    `pos`, it doesn't classify cells as `Likely-cellType`, but strictly annotates 
    cells positive for all defined markers.  
    
    `allneg` : Same as `allpos` but looks for negativity of the defined markers.  
    
    `anypos` : "Anypos" requires only one of the defined markers to be positive. 
    For example, to define macrophages, a cell could be designated as such if 
    any of `CD68`, `CD163`, or `CD206` is positive.  
    
    `anyneg` : Same as `anyneg` but looks for negativity of the defined markers. 
            
## Function
"""

# Library
import numpy as np
import pandas as pd
import anndata as ad
import argparse
import pathlib
import os


# Function
def csPhenotype (csObject,
                    phenotype,
                    midpoint = 0.5,
                    label="phenotype",
                    imageid='imageid',
                    pheno_threshold_percent=None,
                    pheno_threshold_abs=None,
                    fileName=None,
                    verbose=True,
                    projectDir=None):
    """
Parameters:
    csObject (anndata):
        Single or combined CSPOT object.

    phenotype (dataframe, str):
        A phenotyping workflow strategy either as a `dataframe` loaded into memory or 
        a path to the `.csv` file. 

    midpoint (float, optional):
        By default, CSPOT normalizes the data in a way that cells with a value
        above 0.5 are considered positive. However, if you desire more selective
        threshold, the parameter can be adjusted accordingly. 

    label (str, optional):
        Specify the column name under which the final phenotype classification
        will be saved. 

    imageid (str, optional):
        The name of the column that holds the unique image ID. 

    pheno_threshold_percent (float, optional):
        The user-defined threshold, which can be set between 0-100, is used to
        recategorize any phenotype that falls below it as 'unknown'.
        This function is commonly used to address low background false positives.

    pheno_threshold_abs (int, optional):
        This function serves a similar purpose as the `pheno_threshold_percent`,
        but it accepts an absolute number as input. For example, if the user
        inputs 10, any phenotype that contains less than 10 cells will be
        recategorized as 'unknown'. 

    fileName (string, optional):
        File Name to be used while saving the CSPOT object.

    verbose (bool, optional):
        If True, print detailed information about the process to the console.  

    projectDir (string, optional):
        Provide the path to the output directory.
        
Returns:
    csObject (anndata):
        Modified CSPOT object with the Phenotypes is returned. If `projectDir` is 
        provided, it will be saved in the defined directory.

Example:
        ```python
        
        # set the Project directory
        projectDir = '/Users/aj/Documents/cspotExampleData'
        # Path to the CSPOT Object
        csObject = projectDir + '/CSPOT/csObject/exampleImage_cspotPredict.ome.h5ad'
        
        # load the phenotyping workflow
        phenotype = pd.read_csv(str(projectDir) + '/phenotype_workflow.csv')
        
        # Run Function
        adata = cs.csPhenotype ( csObject=csObject,
                            phenotype=phenotype,
                            midpoint = 0.5,
                            label="phenotype",
                            imageid='imageid',
                            pheno_threshold_percent=None,
                            pheno_threshold_abs=None,
                            fileName=None,
                            projectDir=projectDir)
        
        # Same function if the user wants to run it via Command Line Interface
        python csPhenotype.py \
            --csObject /Users/aj/Documents/cspotExampleData/CSPOT/csObject/exampleImage_cspotPredict.ome.h5ad \
            --phenotype /Users/aj/Documents/cspotExampleData/phenotype_workflow.csv \
            --projectDir /Users/aj/Documents/cspotExampleData

        
        ```

    """

    # Load data
    if isinstance(csObject, str):
        adata = ad.read(csObject)
    else:
        adata = csObject.copy()
    
    # load phenotype
    if isinstance(phenotype, pd.DataFrame):
        phenotype = phenotype
    else:
        phenotype = pd.read_csv(pathlib.Path(phenotype))

    # Create a dataframe from the adata object
    data = pd.DataFrame(adata.X, columns = adata.var.index, index= adata.obs.index)

    # Function to calculate the phenotype scores
    def phenotype_cells (data,phenotype,midpoint,group):

        # Subset the phenotype based on the group
        phenotype = phenotype[phenotype.iloc[:,0] == group]

        # Parser to parse the CSV file into four categories
        def phenotype_parser (p, cell):
            # Get the index and subset the phenotype row being passed in
            location = p.iloc[:,1] == cell
            idx = [i for i, x in enumerate(location) if x][0]
            phenotype = p.iloc[idx,:]
            # Calculate
            pos = phenotype[phenotype == 'pos'].index.tolist()
            neg = phenotype[phenotype == 'neg'].index.tolist()
            anypos = phenotype[phenotype == 'anypos'].index.tolist()
            anyneg = phenotype[phenotype == 'anyneg'].index.tolist()
            allpos = phenotype[phenotype == 'allpos'].index.tolist()
            allneg = phenotype[phenotype == 'allneg'].index.tolist()
            return {'pos': pos, 'neg': neg ,'anypos': anypos, 'anyneg': anyneg, 'allpos': allpos, 'allneg': allneg}
            #return pos, neg, anypos, anyneg

        # Run the phenotype_parser function on all rows
        p_list = phenotype.iloc[:,1].tolist()
        r_phenotype = lambda x: phenotype_parser(cell=x, p=phenotype) # Create lamda function
        all_phenotype = list(map(r_phenotype, p_list)) # Apply function
        all_phenotype = dict(zip(p_list, all_phenotype)) # Name the lists

        # Define function to check if there is any marker that does not satisfy the midpoint
        def gate_satisfation_lessthan (marker, data, midpoint):
            fail = np.where(data[marker] < midpoint, 1, 0) # 1 is fail
            return fail
        # Corresponding lamda function
        r_gate_satisfation_lessthan = lambda x: gate_satisfation_lessthan(marker=x, data=data, midpoint=midpoint)

        # Define function to check if there is any marker that does not satisfy the midpoint
        def gate_satisfation_morethan (marker, data, midpoint):
            fail = np.where(data[marker] > midpoint, 1, 0)
            return fail
        # Corresponding lamda function
        r_gate_satisfation_morethan = lambda x: gate_satisfation_morethan(marker=x, data=data, midpoint=midpoint)

        def prob_mapper (data, all_phenotype, cell, midpoint):
            if verbose is True:
                print("Phenotyping " + str(cell))

            # Get the appropriate dict from all_phenotype
            p = all_phenotype[cell]

            # Identiy the marker used in each category
            pos = p.get('pos')
            neg = p.get('neg')
            anypos = p.get('anypos')
            anyneg = p.get('anyneg')
            allpos = p.get('allpos')
            allneg = p.get('allneg')

            # Perform computation for each group independently
            # Positive marker score
            if len(pos) != 0:
                pos_score = data[pos].mean(axis=1).values
                pos_fail = list(map(r_gate_satisfation_lessthan, pos)) if len(pos) > 1 else []
                pos_fail = np.amax(pos_fail, axis=0) if len(pos) > 1 else []
            else:
                pos_score = np.repeat(0, len(data))
                pos_fail = []

            # Negative marker score
            if len(neg) != 0:
                neg_score = (1-data[neg]).mean(axis=1).values
                neg_fail = list(map(r_gate_satisfation_morethan, neg)) if len(neg) > 1 else []
                neg_fail = np.amax(neg_fail, axis=0) if len(neg) > 1 else []
            else:
                neg_score = np.repeat(0, len(data))
                neg_fail = []

            # Any positive score
            anypos_score = np.repeat(0, len(data)) if len(anypos) == 0 else data[anypos].max(axis=1).values

            # Any negative score
            anyneg_score = np.repeat(0, len(data)) if len(anyneg) == 0 else (1-data[anyneg]).max(axis=1).values

            # All positive score
            if len(allpos) != 0:
                allpos_score = data[allpos]
                allpos_score['score'] = allpos_score.max(axis=1)
                allpos_score.loc[(allpos_score < midpoint).any(axis = 1), 'score'] = 0
                allpos_score = allpos_score['score'].values + 0.01 # A small value is added to give an edge over the matching positive cell
            else:
                allpos_score = np.repeat(0, len(data))


            # All negative score
            if len(allneg) != 0:
                allneg_score = 1- data[allneg]
                allneg_score['score'] = allneg_score.max(axis=1)
                allneg_score.loc[(allneg_score < midpoint).any(axis = 1), 'score'] = 0
                allneg_score = allneg_score['score'].values + 0.01
            else:
                allneg_score = np.repeat(0, len(data))


            # Total score calculation
            # Account for differences in the number of categories used for calculation of the final score
            number_of_non_empty_features = np.sum([len(pos) != 0,
                                                len(neg) != 0,
                                                len(anypos) != 0,
                                                len(anyneg) != 0,
                                                len(allpos) != 0,
                                                len(allneg) != 0])

            total_score = (pos_score + neg_score + anypos_score + anyneg_score + allpos_score + allneg_score) / number_of_non_empty_features

            return {cell: total_score, 'pos_fail': pos_fail ,'neg_fail': neg_fail}
            #return total_score, pos_fail, neg_fail


        # Apply the fuction to get the total score for all cell types
        r_prob_mapper = lambda x: prob_mapper (data=data, all_phenotype=all_phenotype, cell=x, midpoint=midpoint) # Create lamda function
        final_scores = list(map(r_prob_mapper, [*all_phenotype])) # Apply function
        final_scores = dict(zip([*all_phenotype], final_scores)) # Name the lists

        # Combine the final score to annotate the cells with a label
        final_score_df = pd.DataFrame()
        for i in [*final_scores]:
            df = pd.DataFrame(final_scores[i][i])
            final_score_df= pd.concat([final_score_df, df], axis=1)
        # Name the columns
        final_score_df.columns = [*final_scores]
        final_score_df.index = data.index
        # Add a column called unknown if all markers have a value less than the midpoint (0.5)
        unknown = group + str('-rest')
        final_score_df[unknown] = (final_score_df < midpoint).all(axis=1).astype(int)

        # Name each cell
        labels = final_score_df.idxmax(axis=1)

        # Group all failed instances (i.e. when multiple markers were given
        # any one of the marker fell into neg or pos zones of the midpoint)
        pos_fail_all = pd.DataFrame()
        for i in [*final_scores]:
            df = pd.DataFrame(final_scores[i]['pos_fail'])
            df.columns = [i] if len(df) != 0 else []
            pos_fail_all= pd.concat([pos_fail_all, df], axis=1)
        pos_fail_all.index = data.index if len(pos_fail_all) != 0 else []
        # Same for Neg
        neg_fail_all = pd.DataFrame()
        for i in [*final_scores]:
            df = pd.DataFrame(final_scores[i]['neg_fail'])
            df.columns = [i] if len(df) != 0 else []
            neg_fail_all= pd.concat([neg_fail_all, df], axis=1)
        neg_fail_all.index = data.index if len(neg_fail_all) != 0 else []


        # Modify the labels with the failed annotations
        if len(pos_fail_all) != 0:
            for i in pos_fail_all.columns:
                labels[(labels == i) & (pos_fail_all[i] == 1)] = 'likely-' + i
        # Do the same for negative
        if len(neg_fail_all) != 0:
            for i in neg_fail_all.columns:
                labels[(labels == i) & (neg_fail_all[i] == 1)] = 'likely-' + i

        # Retun the labels
        return labels

    # Create an empty dataframe to hold the labeles from each group
    phenotype_labels = pd.DataFrame()

    # Loop through the groups to apply the phenotype_cells function
    for i in phenotype.iloc[:,0].unique():

        if phenotype_labels.empty:
            phenotype_labels = pd.DataFrame(phenotype_cells(data = data, group = i, phenotype=phenotype, midpoint=midpoint))
            phenotype_labels.columns = [i]

        else:
            # Find the column with the cell-type of interest
            column_of_interest = [] # Empty list to hold the column name
            try:
                column_of_interest = phenotype_labels.columns[phenotype_labels.eq(i).any()]
            except:
                pass
            # If the cell-type of interest was not found just add NA
            if len(column_of_interest) == 0:
                phenotype_labels[i] = np.nan
            else:
                #cells_of_interest = phenotype_labels[phenotype_labels[column_of_interest] == i].index
                cells_of_interest = phenotype_labels[phenotype_labels[column_of_interest].eq(i).any(axis=1)].index
                d = data.loc[cells_of_interest]
                if verbose is True:
                    print("-- Subsetting " + str(i))
                phenotype_l = pd.DataFrame(phenotype_cells(data = d, group = i, phenotype=phenotype, midpoint=midpoint), columns = [i])
                phenotype_labels = phenotype_labels.merge(phenotype_l, how='outer', left_index=True, right_index=True)

    # Rearrange the rows back to original
    phenotype_labels = phenotype_labels.reindex(data.index)
    phenotype_labels = phenotype_labels.replace('-rest', np.nan, regex=True)
    if verbose is True:
        print("Consolidating the phenotypes across all groups")
        phenotype_labels_Consolidated = phenotype_labels.fillna(method='ffill', axis = 1)
    phenotype_labels[label] = phenotype_labels_Consolidated.iloc[:,-1].values

    # replace nan to 'other cells'
    phenotype_labels[label] = phenotype_labels[label].fillna('Unknown')

    # Apply the phenotype threshold if given
    if pheno_threshold_percent or pheno_threshold_abs is not None:
        p = pd.DataFrame(phenotype_labels[label])
        q = pd.DataFrame(adata.obs[imageid])
        p = q.merge(p, how='outer', left_index=True, right_index=True)

        # Function to remove phenotypes that are less than the given threshold
        def remove_phenotype(p, ID, pheno_threshold_percent, pheno_threshold_abs):
            d = p[p[imageid] == ID]
            x = pd.DataFrame(d.groupby([label]).size())
            x.columns = ['val']
            # FInd the phenotypes that are less than the given threshold
            if pheno_threshold_percent is not None:
                fail = list(x.loc[x['val'] < x['val'].sum() * pheno_threshold_percent/100].index)
            if pheno_threshold_abs is not None:
                fail = list(x.loc[x['val'] < pheno_threshold_abs].index)
            d[label] = d[label].replace(dict(zip(fail, np.repeat('Unknown',len(fail)))))
            # Return
            return d

        # Apply function to all images
        r_remove_phenotype = lambda x: remove_phenotype (p=p, ID=x,
                                                         pheno_threshold_percent=pheno_threshold_percent,
                                                         pheno_threshold_abs=pheno_threshold_abs) # Create lamda function
        final_phrnotypes= list(map(r_remove_phenotype, list(p[imageid].unique()))) # Apply function

        final_phrnotypes = pd.concat(final_phrnotypes, join='outer')
        phenotype_labels = final_phrnotypes.reindex(adata.obs.index)


    # Return to adata
    adata.obs[label] = phenotype_labels[label]

    # Save data if requested
    if projectDir is not None:
        
        finalPath = pathlib.Path(projectDir + '/CSPOT/csPhenotype/')
        if not os.path.exists(finalPath):
            os.makedirs(finalPath)
        # determine file name
        if fileName is None:
            if isinstance (csObject, str):
                imid = pathlib.Path(csObject).stem
            else:
                imid = 'csPhenotype'
        else:
            imid = fileName
    
        adata.write(finalPath / f'{imid}.h5ad')
        # Print
        if verbose is True:
            print('Modified csObject is stored at "' + str(projectDir) + '/CSPOT/csPhenotype')

    else:
        # Return data
        return adata
    return adata


# Make the Function CLI compatable
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='CSPOT Phenotype')
    parser.add_argument('--csObject', type=str, help='Single or combined CSPOT object')
    parser.add_argument('--phenotype', type=str, help='A phenotyping workflow strategy')
    parser.add_argument('--midpoint', type=float, default=0.5, help='Normalization threshold')
    parser.add_argument('--label', type=str, default='phenotype', help='Column name for final phenotype classification')
    parser.add_argument('--imageid', type=str, default='imageid', help='Name of column that holds unique image ID')
    parser.add_argument('--pheno_threshold_percent', type=float, default=None, help='User-defined threshold for recategorizing phenotypes as "unknown"')
    parser.add_argument('--pheno_threshold_abs', type=int, default=None, help='User-defined threshold for recategorizing phenotypes as "unknown"')
    parser.add_argument('--fileName', type=str, default=None, help="File name for saving the modified csObject.")
    parser.add_argument("--verbose", type=bool, default=True, help="If True, print detailed information about the process to the console.")
    parser.add_argument('--projectDir', type=str, help="Directory to save the modified csObject.")
    args = parser.parse_args()
    csPhenotype(csObject=args.csObject,
                   phenotype=args.phenotype,
                   midpoint=args.midpoint,
                   label=args.label,
                   imageid=args.imageid,
                   pheno_threshold_percent=args.pheno_threshold_percent,
                   pheno_threshold_abs=args.pheno_threshold_abs,
                   fileName=args.fileName, 
                   verbose=args.verbose,
                   projectDir=args.projectDir)
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#Created on Sat Feb  4 12:06:32 2023
#@author: Ajit Johnson Nirmal
#Helper function to add the cspotPredicted postive and negative cells to vizualize 
#on the image using scimap. 

"""
!!! abstract "Short Description"
    The `addPredictions` function serves as a link between `cspot` and `scimap` package. 
    It's useful for evaluating model performance. The function transforms results 
    stored in `anndata.uns` to `anndata.obs` so they can be visualized using 
    the `scimap` package's `sm.pl.image viewer` function. This displays `positive` 
    and `negative` cells overlaid on the raw image.
    
    The `addPredictions` function can take in two methods. 
    `cspotOutput` displays the result of running the `cspot` function, 
    while `csScore` shows the raw output produced by the `csScore` 
    function, which returns a probability score. The `midpoint` parameter, 
    with a default value of 0.5, can be adjusted to define what is 
    considered a `positive` result, when method is set to `csScore`.
    
## Function
"""

# libs
import pandas as pd
import anndata as ad
import pathlib
import os
import argparse


# Function
def addPredictions (csObject, 
                    method='cspotOutput',
                    cspotOutput='cspotOutput',
                    csScore='csScore', 
                    midpoint=0.5,
                    outputDir=None):
    """
Parameters:
    csObject (anndata):  
        Single or combined CSPOT object.
        
    method (str, optional):  
        There are two options: `cspotOutput` and `csScore`. 
        `cspotOutput` displays the result of running the `CSPOT` function, 
        while `csScore` shows the raw output produced by the `csScore` 
        function, which returns a probability score. The `midpoint` parameter, 
        with a default value of 0.5, can be adjusted to define what is 
        considered a `positive` result, when method is set to `csScore`.
        
    cspotOutput (str, optional):  
        The name under which the `cspotOutput` is stored.
        
    csScore (str, optional):  
        The name under which the `csScore` is stored.
        
    midpoint (float, optional):  
        The threshold for determining positive cells, in conjunction with 'csScore'.

    outputDir (string, optional):  
        Provide the path to the output directory. Kindly take note that this particular 
        output will not be automatically saved in a predetermined directory, 
        unlike the other outputs. The file will be saved in the directory 
        specified by the `outputDir` parameter. If `None`, the `csObject` will 
        be returned to memory.

Returns:
    csObject (anndata):  
        If output directory is provided the `csObject` will 
        be stored else it will be returned to memory. The results are stored in 
        `anndata.obs` with a `p_` appended to the markers names. So if you would 
        like to vizulaize `CD3`, the column that you are looking for is `p_CD3`.
        
Example:
    	```python    
        
        # Path to projectDir
        projectDir = '/Users/aj/Documents/cspotExampleData'
        
        # Path to csObject
        csObject = projectDir + '/CSPOT/cspotOutput/exampleImage_cspotPredict.ome.h5ad'
        
        adata = cs.addPredictions (csObject, 
                            method='cspotOutput',
                            cspotOutput='cspotOutput',
                            csScore='csScore', 
                            midpoint=0.5)
        
        # Same function if the user wants to run it via Command Line Interface
        python addPredictions.py \
            --csObject Users/aj/Documents/cspotExampleData/CSPOT/cspotOutput/exampleImage_cspotPredict.ome.h5ad    	
        ```

    """
    # Load the adata
    if isinstance(csObject, str):
        adata = ad.read(csObject)
    else: 
        adata = csObject
        
    # function to convert the prob scores to binary pos or neg
    def assign_labels(df, midpoint):
        df = df.applymap(lambda x: 'neg' if x < midpoint else 'pos')
        return df
    
    # intialize the data    
    if method == 'cspotOutput':
        attach_df = adata.uns[cspotOutput].copy()
    elif method == 'csScore':
        df = adata.uns[csScore].copy()
        attach_df = assign_labels (df, midpoint=midpoint)
        
        
    obs = adata.obs.copy()
    columns_to_drop = [col for col in obs.columns if col.startswith('p_')]
    obs.drop(columns_to_drop, axis=1, inplace=True)
    
    new_col_names = ['p_{}'.format(idx) for idx in attach_df.columns]
    attach_df.columns = new_col_names
    # add to obs
    final_obs = pd.concat([obs, attach_df], axis=1)
    adata.obs = final_obs
    
    
    # Return to adata
    # Save data if requested
    if outputDir is not None:    
        finalPath = pathlib.Path(outputDir)     
        if not os.path.exists(finalPath):
            os.makedirs(finalPath)
        # determine file name
        if isinstance (csObject, str):
            imid = pathlib.Path(csObject).stem
        else:
            imid = 'addPredictions'    
        adata.write(finalPath / f'{imid}.h5ad')
    else:
        # Return data
        return adata
    
    

# Make the Function CLI compatable
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Add predictions to an anndata object.')
    parser.add_argument('--csObject', type=str, help='Single or combined CSPOT object.')
    parser.add_argument('--method', type=str, default='cspotOutput', help='Method for determining positive cells (cspotOutput or csScore).')
    parser.add_argument('--cspotOutput', type=str, default='cspotOutput', help='Name under which cspotOutput is stored.')
    parser.add_argument('--csScore', type=str, default='csScore', help='Name under which csScore is stored.')
    parser.add_argument('--midpoint', type=float, default=0.5, help='Threshold for determining positive cells, in conjunction with csScore.')
    parser.add_argument('--outputDir', type=str, default=None, help='Path to the output directory. If None, csObject will be returned to memory.')
    args = parser.parse_args()
    addPredictions(csObject=args.csObject, 
                   method=args.method, 
                   cspotOutput=args.cspotOutput, 
                   csScore=args.csScore, 
                   midpoint=args.midpoint, 
                   outputDir=args.outputDir)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#Created on Wed Dec 28 17:08:09 2022
#@author: Ajit Johnson Nirmal
#Concatenate multiple anndata objects

"""
!!! abstract "Short Description"
    Use `mergecsObject` to combine multiple csObjects into a dataset for 
    analysis when multiple images need to be analyzed.
    
    Note that merging `csObjects` requires merging multiple sections, not 
    simple concatenation. Use parameters to specify which parts of the 
    `csObjects` to merge.
    
## Function
"""

# libs
import pathlib
import pandas as pd
import anndata as ad
import os
import argparse
import numpy as np

# function
def mergecsObject (csObjects,
                      fileName='mergedCSObject',
                      layers=['preProcessed'],
                      uns= ['cspotOutput','csScore','failedMarkers'],
                      verbose=True,
                      projectDir=None):
    """
Parameters:
    csObjects (list):
       A collection of CSPOT Objects to combine into one object, which can
       include both CSPOT Objects stored in memory and those accessed
       via file path.

    fileName (str, optional):
        Designate a Name for the resulting combined CSPOT object.

    layers (list, optional):
        The `.layers` section within the CSPOT Objects to be merged together.

    uns (list, optional):
        The `.uns` section within the CSPOT Objects to be merged together.

    verbose (bool, optional):
        If True, print detailed information about the process to the console. 

    projectDir (str, optional):
        Provide the path to the output directory. The result will be located at
        `projectDir/CSPOT/mergedCSObject/`. 

Returns:
    csObject (anndata):
        If `projectDir` is provided the merged CSPOT Object will saved within the
        provided projectDir.

Example:
        ```python.
        
        # set the working directory & set paths to the example data
        projectDir = '/Users/aj/Documents/cspotExampleData'
        
        csObjects = [projectDir + '/CSPOT/csOutput/exampleImage_cspotPredict.ome.h5ad',
                     projectDir + '/CSPOT/csOutput/exampleImage_cspotPredict.ome.h5ad']
        
        # For this tutorial, supply the same csObject twice for merging, but multiple csObjects can be merged in ideal conditions.
        adata = cs.mergecsObject ( csObjects=csObjects,
                              fileName='mergedcspotObject',
                              layers=['preProcessed'],
                              uns= ['cspotOutput','csScore'],
                              projectDir=projectDir)
        
        # Same function if the user wants to run it via Command Line Interface
        python mergecsObject.py \
            --csObjects /Users/aj/Documents/cspotExampleData/CSPOT/cspotOutput/exampleImage_cspotPredict.ome.h5ad /Users/aj/Documents/cspotExampleData/CSPOT/cspotOutput/exampleImage_cspotPredict.ome.h5ad \
            --projectDir /Users/aj/Documents/cspotExampleData
        
        
        ```
    """

    # Convert to list of anndata objects
    if isinstance (csObjects, list):
        csObjects = csObjects
    else:
        csObjects = [csObjects]

    # converting other parameters to list
    if isinstance (layers, str):
        layers = [layers]
    if isinstance (uns, str):
        uns = [uns]

    # Things to process
    process_layers = ['rawData', 'scaledData', 'obs']
    if layers is not None:
        process_layers.extend(layers)
    if uns is not None:
        process_layers.extend(uns)


    # for expression, uns and layers
    def processX (csObject, process_layers):
        if isinstance(csObject, str):
            # start with raw data
            adata = ad.read(csObject)
        else:
            adata = csObject.copy()

        # print
        if verbose is True:
            print ("Extracting data from: " + str( adata.obs['imageid'].unique()[0]) )

        # process the data
        rawData = pd.DataFrame(adata.raw.X, index=adata.obs.index, columns=adata.var.index)

        # scaled data
        scaledData = pd.DataFrame(adata.X, index=adata.obs.index, columns=adata.var.index)

        # obs
        obs = adata.obs.copy()

        # Process layers
        if layers is not None:
            for i in layers:
                exec(f"{i} = pd.DataFrame(adata.layers[i],index=adata.obs.index, columns=adata.var.index)")

        # Process uns
        if uns is not None:
            for j in uns:
                exec(f"{j} = adata.uns[j]")

        # return the results
        objects = []
        for name in process_layers:
            exec(f"objects.append({name})")

        return objects

    # Run the function
    # Run the function:
    if verbose is True:
        print("Extracting data")
    r_processX = lambda x: processX (csObject=x, process_layers=process_layers)
    processX_result = list(map(r_processX, csObjects)) # Apply function

    # combine all the data
    # create a dictinoary between index and data type
    mapping = {i: element for i, element in enumerate(process_layers)}


    # create an empty dictionary to store the final dataframes
    final_data = {}
    # get the number of lists in the data list
    num_lists = len(processX_result)
    # iterate over the mapping dictionary
    for key, value in mapping.items():
        # create an empty list to store the dataframes
        df_list = []
        # iterate over the data lists
        for i in range(num_lists):
            # retrieve the dataframe from the current list
            df = processX_result[i][key]
            # resolve dict independently
            if isinstance(df, dict):
                df = pd.DataFrame.from_dict(df, orient='index', columns=df[list(df.keys())[0]]).applymap(lambda x: 1)   
            # add the dataframe to the df_list
            df_list.append(df)
        # concatenate the dataframes in the df_list
        df = pd.concat(df_list)
        # add the resulting dataframe to the final_data dictionary
        final_data[value] = df


    # create the combined anndata object
    bdata = ad.AnnData(final_data.get("rawData"), dtype=np.float64)
    bdata.obs = final_data.get("obs")
    bdata.raw = bdata
    bdata.X = final_data.get("scaledData")
    # add layers
    if layers is not None:
        for i in layers:
            tmp = final_data.get(i)
            bdata.layers[i] = tmp
    # add uns
    if uns is not None:
        for i in uns:
            tmp = final_data.get(i)
            bdata.uns[i] = tmp

    # last resolve all_markers if it exisits
    if isinstance(csObjects[0], str):
        # start with raw data
        adata = ad.read(csObjects[0])
    else:
        adata = csObjects[0].copy()
    if hasattr(adata, 'uns') and 'all_markers' in adata.uns:
        # the call to adata.uns['all_markers'] is valid
        bdata.uns['all_markers'] = adata.uns['all_markers']

    # write the output
    if projectDir is not None:
        finalPath = pathlib.Path(projectDir + '/CSPOT/mergedcsObject')
        if not os.path.exists(finalPath):
            os.makedirs(finalPath)
        bdata.write(finalPath / f'{fileName}.h5ad')
        # Print
        if verbose is True:
            print('Given csObjects have been merged, head over to "' + str(projectDir) + '/CSPOT/mergedcsObject" to view results')


    # return data
    return bdata

# Make the Function CLI compatable
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Merge CSPOT Objects')
    parser.add_argument('--csObjects', type=str, nargs='+', help='A collection of CSPOT Objects to combine into one object')
    parser.add_argument('--fileName', type=str, default='mergedcsObject', help='Designate a Name for the resulting combined CSPOT object')
    parser.add_argument('--layers', type=str, nargs='+', default=['preProcessed'], help='The layers section within the CSPOT Objects to be merged together')
    parser.add_argument('--uns', type=str, nargs='+', default=['cspotOutput','csScore'], help='The uns section within the CSPOT Objects to be merged together')
    parser.add_argument("--verbose", type=bool, default=True, help="If True, print detailed information about the process to the console.")       
    parser.add_argument('--projectDir', type=str, default=None, help='Provide the path to the output directory')
    args = parser.parse_args()
    mergecsObject(csObjects=args.csObjects,
                     fileName=args.fileName,
                     layers=args.layers, 
                     uns=args.uns,
                     verbose=args.verbose,
                     projectDir=args.projectDir)




"""
!!! abstract "Short Description"
    The function `csPredict` is employed to make predictions about the 
    expression of a specified marker on cells in new images using the models 
    generated by `csTrain`. This calculation is done at the pixel level, 
    resulting in an output image where the number of channels corresponds to 
    the number of models applied to the input image. The parameter `markerChannelMapPath` 
    is used to associate the image channel number with the relevant model to be applied.


## Function
"""


# Libs
import os, argparse
import pandas as pd
import pathlib

# tools libs
from skimage import io as skio
import tensorflow.compat.v1 as tf
import tifffile
import numpy as np
from skimage.transform import resize


# from other .py scripts
if __name__ == '__main__':
    from toolbox.imtools import im2double
    from UNet import UNet2D
elif __name__ == 'csPredict':
    from toolbox.imtools import im2double
    from UNet import UNet2D
else:
    from .toolbox.imtools import im2double
    from .UNet import UNet2D



# Function
def csPredict (imagePath,
                 csModelPath,
                 projectDir, 
                 markerChannelMapPath, 
                 markerColumnName='marker', 
                 channelColumnName='channel', 
                 modelColumnName='cspotmodel', 
                 verbose=True,
                 GPU=-1,
                 dsFactor=1):
    
    """
Parameters:
    imagePath (str):  
        The path to the .tif file that needs to be processed. 
     
    csModelPath (str):  
        The path to the `cspotModel` folder. 
    
    projectDir (str):  
        The path to the output directory where the processed images (`probabilityMasks`) will be saved.
     
    markerChannelMapPath (str):  
        The path to the marker panel list, which contains information about the markers used in the image.
     
    markerColumnName (str, optional):  
        The name of the column in the marker panel list that contains the marker names. 
     
    channelColumnName (str, optional):  
        The name of the column in the marker panel list that contains the channel names. 
     
    modelColumnName (str, optional):  
        The name of the column in the marker panel list that contains the model names. 

    verbose (bool, optional):
        If True, print detailed information about the process to the console.  

    GPU (int, optional):  
        An optional argument to explicitly select the GPU to use. The default value is -1, meaning that the GPU will be selected automatically.

    dsFactor (float, optional):
        An optional argument to downsample image before inference. The default value is 1, meaning that the image is not downsampled. Use it to modify image pixel size to match training data in the model.

Returns:
    Predicted Probability Masks (images):  
        The result will be located at `projectDir/CSPOT/csPredict/`.

Example:
    	```python    
        # Path to all the files that are necessary files for running csPredict
        projectDir = '/Users/aj/Documents/cspotExampleData'
        
        # csPredict related paths
        imagePath = projectDir + '/image/exampleImage.tif'
        markerChannelMapPath = projectDir + '/markers.csv'
        csModelPath = projectDir + '/manuscriptModels/'
        
        # Run the function
        cs.csPredict( imagePath=imagePath,
                 csModelPath=csModelPath,
                 projectDir=projectDir,
                 markerChannelMapPath=markerChannelMapPath, 
                 markerColumnName='marker', 
                 channelColumnName='channel', 
                 modelColumnName='cspotmodel')
        
        # Same function if the user wants to run it via Command Line Interface
        python csPredict.py \
            --imagePath /Users/aj/Documents/cspotExampleData/image/exampleImage.tif \
            --csModelPath /Users/aj/Documents/cspotExampleData/manuscriptModels \
            --projectDir /Users/aj/Documents/cspotExampleData \
            --markerChannelMapPath /Users/aj/Documents/cspotExampleData/markers.csv
    	
    	```
     
     """
    
    fileName = pathlib.Path(imagePath).stem

    # read the markers.csv
    maper = pd.read_csv(pathlib.Path(markerChannelMapPath))
    columnnames =  [word.lower() for word in maper.columns]
    maper.columns = columnnames
    
    # making it compatable with mcmicro when no channel info is provided
    if not set(['channel', 'channels', channelColumnName]).intersection(set(columnnames)):
        # add a column called 'channel'
        maper['channel'] = [i + 1 for i in range(len(maper))]
        columnnames = list(maper.columns)
        
        
    # identify the marker column name (doing this to make it easier for people who confuse between marker and markers)
    if markerColumnName not in columnnames:
        # ckeck if 'markers' or 'marker_name' or 'marker_names' is in columnnames
        # if so assign that match to markerCol
        for colname in columnnames:
            if 'marker' in colname or 'markers' in colname or 'marker_name' in colname or 'marker_names' in colname:
                markerCol = colname
                break
        else:
            raise ValueError('markerColumnName not found in markerChannelMap, please check')
    else:
        markerCol = markerColumnName


   # identify the channel column name (doing this to make it easier for people who confuse between channel and channels)
    if channelColumnName not in columnnames:
        if channelColumnName != 'channel':
            raise ValueError('channelColumnName not found in markerChannelMap, please check')
        if 'channels' in columnnames:
            channelCol = 'channels'
        else:
            raise ValueError('channelColumnName not found in markerChannelMap, please check')
    else:
        channelCol = channelColumnName


    # identify the CSPOT model column name (doing this to make it easier for people who confuse between cspotmodel and cspotmodels)
    if modelColumnName not in columnnames:
        if modelColumnName != 'cspotmodel':
            raise ValueError('modelColumnName not found in markerChannelMap, please check')
        if 'cspotmodels' in columnnames:
            modelCol = 'cspotmodels'
        else:
            raise ValueError('modelColumnName not found in markerChannelMap, please check')
    else:
        modelCol = modelColumnName

    # remove rowa that have nans in modelCol
    runMenu = maper.dropna(subset=[modelCol], inplace=False)[[channelCol,markerCol,modelCol]]

    # shortcuts
    numMarkers = len(runMenu)

    I = skio.imread(imagePath, img_num=0, plugin='tifffile')


    probPath = pathlib.Path(projectDir + '/CSPOT/csPredict/')
    modelPath = pathlib.Path(csModelPath)

    if not os.path.exists(probPath):
        os.makedirs(probPath,exist_ok=True)


    def data(runMenu, 
             imagePath, 
             modelPath, 
             projectDir, 
             dsFactor=dsFactor, 
             GPU=GPU):
        
        # Loop through the rows of the DataFrame
        for index, row in runMenu.iterrows():
            channel = row[channelCol]
            markerName = row[markerCol]
            cspotmodel = row[modelCol]
            if verbose is True:
                print('Running CSPOT model ' + str(cspotmodel) + ' on channel ' + str(channel) + ' corresponding to marker ' + str(markerName) )


            tf.reset_default_graph()
            UNet2D.singleImageInferenceSetup(pathlib.Path(modelPath / cspotmodel), GPU, -1, -1)

            fileName = os.path.basename(imagePath)
            fileNamePrefix = fileName.split(os.extsep, 1)
            fileType = fileNamePrefix[1]
            if fileType == 'ome.tif' or fileType == 'ome.tiff' or fileType == 'btf':
                I = skio.imread(imagePath, img_num=int(channel-1), plugin='tifffile')
            elif fileType == 'tif':
                I = tifffile.imread(imagePath, key=int(channel-1))

            if I.dtype == 'float32':
                I = im2double(I) * 255
            elif I.dtype == 'uint16':
                I = im2double(I) * 255

            rawVert = I.shape[0]
            rawHorz = I.shape[1]
            hsize = int(float(rawVert * float(dsFactor)))
            vsize = int(float(rawHorz * float(dsFactor)))
            I = resize(I, (hsize, vsize),preserve_range=True)

            append_kwargs = {
                'bigtiff': True,
                'metadata': None,
                'append': True,
            }
            save_kwargs = {
                'bigtiff': True,
                'metadata': None,
                'append': False,
            }

            PM = np.uint8(255 * UNet2D.singleImageInference(I, 'accumulate',1))
            PM = resize(PM, (rawVert, rawHorz))
            yield np.uint8(255 * PM)

    with tifffile.TiffWriter(probPath / (fileName + '_cspotPredict.ome.tif'), bigtiff=True) as tiff:
        tiff.write(data(runMenu, imagePath, modelPath, probPath, dsFactor=dsFactor, GPU=GPU), shape=(numMarkers,I.shape[0],I.shape[1]), dtype='uint8', metadata={'Channel': {'Name': runMenu[markerCol].tolist()}, 'axes': 'CYX'})
        UNet2D.singleImageInferenceCleanup()


# =============================================================================
#     logPath = ''
#     scriptPath = os.path.dirname(os.path.realpath(__file__))
# 
#     pmPath = ''
# 
#     if os.system('nvidia-smi') == 0:
#         if args.GPU == -1:
#             print("automatically choosing GPU")
#             GPU = GPUselect.pick_gpu_lowest_memory()
#         else:
#             GPU = args.GPU
#         print('Using GPU ' + str(GPU))
# 
#     else:
#         if sys.platform == 'win32':  # only 1 gpu on windows
#             if args.GPU == -1:
#                 GPU = 0
#                 print('using default GPU')
#             else:
#                 GPU = args.GPU
#             print('Using GPU ' + str(GPU))
#         else:
#             GPU = 0
#             print('Using CPU')
#     os.environ['CUDA_VISIBLE_DEVICES'] = '%d' % GPU
# =============================================================================


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='CSPOT Predict function')
    parser.add_argument('--imagePath', type=str, required=True, help='The path to the .tif file that needs to be processed.')
    parser.add_argument('--csModelPath', type=str, required=True, help='The path to the `cspotModel` folder.')
    parser.add_argument('--projectDir', type=str, help='The path to the output directory where the processed images will be saved.')
    parser.add_argument('--markerChannelMapPath', type=str, required=True, help='The path to the marker panel list, which contains information about the markers used in the image.')
    parser.add_argument('--markerColumnName', type=str, default='marker', help='The name of the column in the marker panel list that contains the marker names. The default value is `marker`.')
    parser.add_argument('--channelColumnName', type=str, default='channel', help='The name of the column in the marker panel list that contains the channel names. The default value is `channel`.')
    parser.add_argument('--modelColumnName', type=str, default='cspotmodel', help='The name of the column in the marker panel list that contains the model names. The default value is `cspotmodel`.')
    parser.add_argument("--verbose", type=bool, default=True, help="If True, print detailed information about the process to the console.")      
    parser.add_argument('--GPU', type=int, default=-1, help='An optional argument to explicitly select the GPU to use. The default value is 0, meaning that the GPU will be selected automatically.')
    parser.add_argument('--dsFactor', type=float, default=1,help='Scaling Factor.')

    args = parser.parse_args()
    csPredict (imagePath=args.imagePath, 
                 csModelPath=args.csModelPath, 
                 projectDir=args.projectDir, 
                 markerChannelMapPath=args.markerChannelMapPath, 
                 markerColumnName=args.markerColumnName, 
                 channelColumnName=args.channelColumnName, 
                 modelColumnName=args.modelColumnName, 
                 verbose=args.verbose,
                 GPU=args.GPU,
                 dsFactor=args.dsFactor)









# -*- coding: utf-8 -*-
#Created on Fri Mar 17 10:00:18 2023
#@author: Ajit Johnson Nirmal
#Plotting function to visualize postive cells

"""
!!! abstract "Short Description"
    The scatterPlot function can be used to create scatter plots of single-cell spatial data. 
    This function can be used to visualize the spatial distribution of positive 
    cells for a given marker, providing a quick and intuitive way to view the final predictions.

## Function
"""

# Libs
import anndata as ad
import pathlib
import matplotlib.pyplot as plt
import argparse
import matplotlib as mpl
mpl.rcParams['pdf.fonttype'] = 42

# Function
def scatterPlot (csObject, 
                 markers=None, 
                 cspotOutput='cspotOutput',
                 x_coordinate='X_centroid',
                 y_coordinate='Y_centroid',
                 poscellsColor='#78290f',
                 negcellsColor='#e5e5e5',
                 s=None,
                 ncols=5,
                 alpha=1,
                 dpi=200,
                 figsize=(5, 5),
                 invert_yaxis=True,
                 outputDir=None,
                 outputFileName='cspotPlot.png',
                 **kwargs):
    """
Parameters:
    csObject (anndata):
        Pass the `csObject` loaded into memory or a path to the `csObject` 
        file (.h5ad).
        
    markers (str or list of str, optional): 
        The name(s) of the markers to plot. If not provided, all markers will be plotted.
        
    cspotOutput (str, optional): 
        The label underwhich the CSPOT output is stored within the object.
        
    x_coordinate (str, optional):
        The column name in `single-cell spatial table` that records the
        X coordinates for each cell. 

    y_coordinate (str, optional):
        The column name in `single-cell spatial table` that records the
        Y coordinates for each cell.
        
    poscellsColor (str, optional): 
        The color of positive cells.
        
    negcellsColor (str, optional): 
        The color of negative cells. 
        
    s (float, optional): 
        The size of the markers.
        
    ncols (int, optional): 
        The number of columns in the final plot when multiple makers are plotted. 
        
    alpha (float, optional): 
        The alpha value of the points (controls opacity).
        
    dpi (int, optional): 
        The DPI of the figure.
        
    figsize (tuple, optional): 
        The size of the figure.

    invert_yaxis (bool, optional):  
        Invert the Y-axis of the plot. 
        
    outputDir (str, optional): 
        The directory to save the output plot. 
        
    outputFileName (str, optional): 
        The name of the output file. Use desired file format as 
        suffix (e.g. `.png` pr `.pdf`)

    **kwargs (keyword parameters):
        Additional arguments to pass to the `matplotlib.scatter` function.


Returns:
    Plot (image):
        If `outputDir` is provided the plot will saved within the
        provided outputDir.

Example:
        ```python
        
        # Prohect directory
        projectDir = '/Users/aj/Documents/cspotExampleData'
        
        # path to the final output
        csObject = '/Users/aj/Desktop/cspotExampleData/CSPOT/cspotOutput/exampleImage_cspotPredict.ome.h5ad'
        
        # Plot image to console
        cs.scatterPlot(csObject,
            markers=['ECAD', 'CD8A', 'CD45'],
            poscellsColor='#78290f',
            negcellsColor='#e5e5e5',
            s=3,
            ncols=3,
            dpi=90,
            figsize=(4, 4),
            outputDir=None,
            outputFileName='cspotplot.png')
        
        # Same function if the user wants to run it via Command Line Interface
        python scatterPlot.py --csObject /Users/aj/Desktop/cspotExampleData/CSPOT/cspotOutput/exampleImage_cspotPredict.ome.h5ad \
                            --markers ECAD CD8A \
                            --outputDir /Users/aj/Desktop/cspotExampleData/CSPOT
        ```
        
    """
    
    # Load the andata object
    if isinstance(csObject, str):
        adata = ad.read(csObject)
    else:
        adata = csObject.copy()
        
    # break the function if cspotOutput is not detectable
    def check_key_exists(dictionary, key):
        try:
            # Check if the key exists in the dictionary
            value = dictionary[key]
        except KeyError:
            # Return an error if the key does not exist
            return "Error: " + str(cspotOutput) + " does not exist, please check!"
    # Test
    check_key_exists(dictionary=adata.uns, key=cspotOutput)

    
    # convert marter to list
    if markers is None:
        markers = list(adata.uns[cspotOutput].columns)
    if isinstance (markers, str):
        markers = [markers]
        
    # identify the x and y coordinates
    x = adata.obs[x_coordinate]
    y = adata.obs[y_coordinate]
    
    # subset the cspotOutput with the requested markers
    subset = adata.uns[cspotOutput][markers]
    # get the list of columns to plot
    cols_to_plot = subset.columns
    
    # identify the number of columns to plot
    ncols = min(ncols, len(cols_to_plot))
    
    # calculate the number of rows needed for the subplot
    nrows = (len(cols_to_plot) - 1) // ncols + 1
    
    # resolve figsize
    figsize = (figsize[0]*ncols, figsize[1]*nrows)
    
    # Estimate point size
    if s is None:
        s = (100000 / adata.shape[0]) / len(cols_to_plot)
        
    # FIIGURE
    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, dpi=dpi)
    for i, col in enumerate(cols_to_plot):
        # get the classes for the current column
        classes = list(subset[col])
        
        # get the current subplot axes
        if nrows==1 and ncols==1:
            ax = axs
        elif nrows==1 or ncols==1:
            ax = axs[i]
        else:
            ax = axs[i // ncols, i % ncols]
        
        # invert y-axis
        if invert_yaxis is True:
            ax.invert_yaxis()
        
        # set the title of the subplot to the current column name
        ax.set_title(col)
        
        # plot the 'neg' points with a small size
        neg_x = [x[j] for j in range(len(classes)) if classes[j] == 'neg']
        neg_y = [y[j] for j in range(len(classes)) if classes[j] == 'neg']
        #ax.scatter(x=neg_x, y=neg_y, c=negcellsColor, s=s, alpha=alpha)
        ax.scatter(x=neg_x, y=neg_y, c=negcellsColor, s=s, linewidth=0, alpha=alpha, **kwargs)
    
        # plot the 'pos' points on top of the 'neg' points with a larger size
        pos_x = [x[j] for j in range(len(classes)) if classes[j] == 'pos']
        pos_y = [y[j] for j in range(len(classes)) if classes[j] == 'pos']
        #ax.scatter(x=pos_x, y=pos_y, c=poscellsColor, s=s, alpha=alpha)
        ax.scatter(x=pos_x, y=pos_y, c=poscellsColor, s=s, linewidth=0, alpha=alpha, **kwargs)
    
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_xticks([])
        ax.set_yticks([])
        
    # remove any unused subplots
    for i in range(len(cols_to_plot), nrows * ncols):
        fig.delaxes(axs[i // ncols, i % ncols])
    
    plt.tight_layout()

    # save figure
    if outputDir is not None:
        plt.savefig(pathlib.Path(outputDir) / outputFileName)
    
    #plt.show()
    

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Create a scatter plot of cells from a CSPOT object.')
    parser.add_argument('--csObject', type=str, help='Path to the csObject file or name of the variable in memory.')
    parser.add_argument('--markers', type=str, nargs='+', default=None, help='The name(s) of the markers to plot. If not provided, all markers will be plotted.')
    parser.add_argument('--cspotOutput', type=str, default='cspotOutput', help='The label underwhich the CSPOT output is stored within the object.')
    parser.add_argument('--x_coordinate', type=str, default='X_centroid', help='The column name in `single-cell spatial table` that records the X coordinates for each cell.')
    parser.add_argument('--y_coordinate', type=str, default='Y_centroid', help='The column name in `single-cell spatial table` that records the Y coordinates for each cell.')
    parser.add_argument('--poscellsColor', type=str, default='#78290f', help='The color of positive cells.')
    parser.add_argument('--negcellsColor', type=str, default='#e5e5e5', help='The color of negative cells.')
    parser.add_argument('--s', type=float, default=None, help='The size of the markers.')
    parser.add_argument('--ncols', type=int, default=5, help='The number of columns in the final plot when multiple markers are plotted.')
    parser.add_argument('--alpha', type=float, default=1, help='The alpha value of the points (controls opacity).')
    parser.add_argument('--dpi', type=int, default=200, help='The DPI of the figure.')
    parser.add_argument('--figsize', type=float, nargs=2, default=[5, 5], help='The size of the figure.')
    parser.add_argument('--outputDir', type=str, default=None, help='The directory to save the output plot.')
    parser.add_argument('--outputFileName', type=str, default='cspotPlot.png', help='The name of the output file. Use desired file format as suffix (e.g. `.png` or `.pdf`).')
    args = parser.parse_args()
    
    # Call the function with the argparse arguments as parameters
    scatterPlot(csObject=args.csObject,
                markers=args.markers,
                cspotOutput=args.cspotOutput,
                x_coordinate=args.x_coordinate,
                y_coordinate=args.y_coordinate,
                poscellsColor=args.poscellsColor,
                negcellsColor=args.negcellsColor,
                s=args.s,
                ncols=args.ncols,
                alpha=args.alpha,
                dpi=args.dpi,
                figsize=tuple(args.figsize),
                outputDir=args.outputDir,
                outputFileName=args.outputFileName)




"""
split/merge utilities to help processing large images; 'object masks' version
"""

import numpy as np
from toolbox.toolbox import *

class PI2D:
    Image = None
    SuggestedPatchSize = 128
    Margin = 14
    PC = None # patch coordinates
    NumPatches = 0
    Output = None
    NR = None
    NC = None
    Boxes = []
    Contours = []
    OutputRaw = None
    Output = None

    def setup(image,suggestedPatchSize,margin):
        """
        initialize PI2D

        *inputs:*
            image: 2D image to partition; assumed double, single channel, in range [0, 1]

            suggestedPatchSize: suggested size of square patch (tile);
            actual patch sizes may vary depending on image size

            margin: half the amount of overlap between adjacent patches; margin should be
            an integer greater than 0 and smaller than suggestedPatchSize/2
        """

        PI2D.Boxes = []
        PI2D.Contours = []

        PI2D.Image = image
        PI2D.SuggestedPatchSize = suggestedPatchSize
        PI2D.Margin = margin

        if len(image.shape) == 2:
            nr,nc = image.shape
        elif len(image.shape) == 3: # multi-channel image
            nz,nr,nc = image.shape

        PI2D.NR = nr
        PI2D.NC = nc

        npr = int(np.ceil(nr/suggestedPatchSize)) # number of patch rows
        npc = int(np.ceil(nc/suggestedPatchSize)) # number of patch cols

        pcRows = np.linspace(0, nr, npr+1).astype(int)
        pcCols = np.linspace(0, nc, npc+1).astype(int)

        PI2D.PC = [] # patch coordinates [r0,r1,c0,c1]
        for i in range(npr):
            r0 = np.maximum(pcRows[i]-margin, 0)
            r1 = np.minimum(pcRows[i+1]+margin, nr)
            for j in range(npc):
                c0 = np.maximum(pcCols[j]-margin, 0)
                c1 = np.minimum(pcCols[j+1]+margin, nc)
                PI2D.PC.append([r0,r1,c0,c1])

        PI2D.NumPatches = len(PI2D.PC)

        PI2D.OutputRaw = 0.25*PI2D.Image
        PI2D.Output = np.copy(PI2D.OutputRaw)

    def getPatch(i):
        """
        returns the i-th patch for processing
        """

        r0,r1,c0,c1 = PI2D.PC[i]
        if len(PI2D.Image.shape) == 2:
            return PI2D.Image[r0:r1,c0:c1]
        if len(PI2D.Image.shape) == 3:
            return PI2D.Image[:,r0:r1,c0:c1]

    def patchOutput(i,bbs,cts):
        """
        adds result bounding boxes (bbs) and countours (cts)
        of i-th tile processing to the output image
        """

        r0,r1,c0,c1 = PI2D.PC[i]
        for idx in range(len(bbs)):
            xmin, ymin, xmax, ymax = bbs[idx] # x: cols; y: rows
            ct = cts[idx]#np.array(cts[idx])
            
            for row in range(ymin,ymax+1):
                PI2D.OutputRaw[r0+row, c0+xmin] = 0.5
                PI2D.OutputRaw[r0+row, c0+xmax] = 0.5
            for col in range(xmin, xmax+1):
                PI2D.OutputRaw[r0+ymin, c0+col] = 0.5
                PI2D.OutputRaw[r0+ymax, c0+col] = 0.5
            for rc in ct:
                PI2D.OutputRaw[r0+rc[0],c0+rc[1]] = 1

            xmin += c0
            xmax += c0
            ymin += r0
            ymax += r0
            ct[:,0] += r0
            ct[:,1] += c0
            
            candidate_box = [xmin, ymin, xmax, ymax]
            candidate_contour = ct

            if PI2D.Boxes:
                did_find_redundancy = False
                for index_box in range(len(PI2D.Boxes)):
                    box = PI2D.Boxes[index_box]
                    if boxes_intersect(candidate_box, box):
                        contour = PI2D.Contours[index_box]

                        cc = np.concatenate((candidate_contour, contour), axis=0)
                        cc_min_r, cc_min_c = np.min(cc, axis=0)
                        cc_max_r, cc_max_c = np.max(cc, axis=0)

                        cc_box_a = np.zeros((cc_max_r-cc_min_r+1, cc_max_c-cc_min_c+1), dtype=bool)
                        cc_box_b = np.copy(cc_box_a)

                        for idx_c in range(candidate_contour.shape[0]):
                            cc_box_a[candidate_contour[idx_c,0]-cc_min_r,candidate_contour[idx_c,1]-cc_min_c] = True

                        for idx_c in range(contour.shape[0]):
                            cc_box_b[contour[idx_c,0]-cc_min_r,contour[idx_c,1]-cc_min_c] = True

                        cc_box_a = imfillholes(cc_box_a)
                        cc_box_b = imfillholes(cc_box_b)

                        if np.sum(cc_box_a*cc_box_b) > 0:
                            candidate_area = np.sum(cc_box_a)
                            area = np.sum(cc_box_b)
                            if candidate_area > area:
                                PI2D.Boxes[index_box] = candidate_box
                                PI2D.Contours[index_box] = candidate_contour
                            did_find_redundancy = True
                            break
                if not did_find_redundancy:
                    PI2D.Boxes.append(candidate_box)
                    PI2D.Contours.append(candidate_contour)
            else:
                PI2D.Boxes.append(candidate_box)
                PI2D.Contours.append(candidate_contour)

    def prepareOutput():
        """
        computes output with resolved intersections in overlapping areas
        which is accessible at PI2D.Output; the output with unresolved
        intersections is accessible at PI2D.OutputRaw
        """

        boxes = PI2D.Boxes
        contours = PI2D.Contours
        PI2D.OutputBoxes = np.copy(PI2D.OutputRaw) * 0
        PI2D.Outputlabel = np.copy(PI2D.Image) * 0
        for idx in range(len(boxes)):
            xmin, ymin, xmax, ymax = boxes[idx] # x: cols; y: rows
            ct = contours[idx]
            for row in range(ymin,ymax+1):
                PI2D.OutputBoxes[row, xmin] = 1
                PI2D.OutputBoxes[row, xmax] = 1
            for col in range(xmin, xmax+1):
                PI2D.OutputBoxes[ymin, col] = 1
                PI2D.OutputBoxes[ymax, col] = 1
            for rc in ct:
                PI2D.Outputlabel[rc[0],rc[1]] = 1



import numpy as np
#from toolbox.imtools import *
# from toolbox.ftools import *
# import sys

class PI2D:
    Image = None
    PaddedImage = None
    PatchSize = 128
    Margin = 14
    SubPatchSize = 100
    PC = None # patch coordinates
    NumPatches = 0
    Output = None
    Count = None
    NR = None
    NC = None
    NRPI = None
    NCPI = None
    Mode = None
    W = None

    def setup(image,patchSize,margin,mode):
        PI2D.Image = image
        PI2D.PatchSize = patchSize
        PI2D.Margin = margin
        subPatchSize = patchSize-2*margin
        PI2D.SubPatchSize = subPatchSize

        W = np.ones((patchSize,patchSize))
        W[[0,-1],:] = 0
        W[:,[0,-1]] = 0
        for i in range(1,2*margin):
            v = i/(2*margin)
            W[i,i:-i] = v
            W[-i-1,i:-i] = v
            W[i:-i,i] = v
            W[i:-i,-i-1] = v
        PI2D.W = W

        if len(image.shape) == 2:
            nr,nc = image.shape
        elif len(image.shape) == 3: # multi-channel image
            nz,nr,nc = image.shape

        PI2D.NR = nr
        PI2D.NC = nc

        npr = int(np.ceil(nr/subPatchSize)) # number of patch rows
        npc = int(np.ceil(nc/subPatchSize)) # number of patch cols

        nrpi = npr*subPatchSize+2*margin # number of rows in padded image 
        ncpi = npc*subPatchSize+2*margin # number of cols in padded image 

        PI2D.NRPI = nrpi
        PI2D.NCPI = ncpi

        if len(image.shape) == 2:
            PI2D.PaddedImage = np.zeros((nrpi,ncpi))
            PI2D.PaddedImage[margin:margin+nr,margin:margin+nc] = image
        elif len(image.shape) == 3:
            PI2D.PaddedImage = np.zeros((nz,nrpi,ncpi))
            PI2D.PaddedImage[:,margin:margin+nr,margin:margin+nc] = image

        PI2D.PC = [] # patch coordinates [r0,r1,c0,c1]
        for i in range(npr):
            r0 = i*subPatchSize
            r1 = r0+patchSize
            for j in range(npc):
                c0 = j*subPatchSize
                c1 = c0+patchSize
                PI2D.PC.append([r0,r1,c0,c1])

        PI2D.NumPatches = len(PI2D.PC)
        PI2D.Mode = mode # 'replace' or 'accumulate'

    def getPatch(i):
        r0,r1,c0,c1 = PI2D.PC[i]
        if len(PI2D.PaddedImage.shape) == 2:
            return PI2D.PaddedImage[r0:r1,c0:c1]
        if len(PI2D.PaddedImage.shape) == 3:
            return PI2D.PaddedImage[:,r0:r1,c0:c1]

    def createOutput(nChannels):
        if nChannels == 1:
            PI2D.Output = np.zeros((PI2D.NRPI,PI2D.NCPI),np.float16)
        else:
            PI2D.Output = np.zeros((nChannels,PI2D.NRPI,PI2D.NCPI),np.float16)
        if PI2D.Mode == 'accumulate':
            PI2D.Count = np.zeros((PI2D.NRPI,PI2D.NCPI),np.float16)

    def patchOutput(i,P):
        r0,r1,c0,c1 = PI2D.PC[i]
        if PI2D.Mode == 'accumulate':
            PI2D.Count[r0:r1,c0:c1] += PI2D.W
        if len(P.shape) == 2:
            if PI2D.Mode == 'accumulate':
                PI2D.Output[r0:r1,c0:c1] += np.multiply(P,PI2D.W)
            elif PI2D.Mode == 'replace':
                PI2D.Output[r0:r1,c0:c1] = P
        elif len(P.shape) == 3:
            if PI2D.Mode == 'accumulate':
                for i in range(P.shape[0]):
                    PI2D.Output[i,r0:r1,c0:c1] += np.multiply(P[i,:,:],PI2D.W)
            elif PI2D.Mode == 'replace':
                PI2D.Output[:,r0:r1,c0:c1] = P

    def getValidOutput():
        margin = PI2D.Margin
        nr, nc = PI2D.NR, PI2D.NC
        if PI2D.Mode == 'accumulate':
            C = PI2D.Count[margin:margin+nr,margin:margin+nc]
        if len(PI2D.Output.shape) == 2:
            if PI2D.Mode == 'accumulate':
                return np.divide(PI2D.Output[margin:margin+nr,margin:margin+nc],C)
            if PI2D.Mode == 'replace':
                return PI2D.Output[margin:margin+nr,margin:margin+nc]
        if len(PI2D.Output.shape) == 3:
            if PI2D.Mode == 'accumulate':
                for i in range(PI2D.Output.shape[0]):
                    PI2D.Output[i,margin:margin+nr,margin:margin+nc] = np.divide(PI2D.Output[i,margin:margin+nr,margin:margin+nc],C)
            return PI2D.Output[:,margin:margin+nr,margin:margin+nc]


    def demo():
        I = np.random.rand(128,128)
        # PI2D.setup(I,128,14)
        PI2D.setup(I,64,4,'replace')

        nChannels = 2
        PI2D.createOutput(nChannels)

        for i in range(PI2D.NumPatches):
            P = PI2D.getPatch(i)
            Q = np.zeros((nChannels,P.shape[0],P.shape[1]))
            for j in range(nChannels):
                Q[j,:,:] = P
            PI2D.patchOutput(i,Q)

        J = PI2D.getValidOutput()
        J = J[0,:,:]

        D = np.abs(I-J)
        print(np.max(D))

        K = cat(1,cat(1,I,J),D)
        imshow(K)


class PI3D:
    Image = None
    PaddedImage = None
    PatchSize = 128
    Margin = 14
    SubPatchSize = 100
    PC = None # patch coordinates
    NumPatches = 0
    Output = None
    Count = None
    NR = None # rows
    NC = None # cols
    NZ = None # planes
    NRPI = None
    NCPI = None
    NZPI = None
    Mode = None
    W = None

    def setup(image,patchSize,margin,mode):
        PI3D.Image = image
        PI3D.PatchSize = patchSize
        PI3D.Margin = margin
        subPatchSize = patchSize-2*margin
        PI3D.SubPatchSize = subPatchSize

        W = np.ones((patchSize,patchSize,patchSize))
        W[[0,-1],:,:] = 0
        W[:,[0,-1],:] = 0
        W[:,:,[0,-1]] = 0
        for i in range(1,2*margin):
            v = i/(2*margin)
            W[[i,-i-1],i:-i,i:-i] = v
            W[i:-i,[i,-i-1],i:-i] = v
            W[i:-i,i:-i,[i,-i-1]] = v

        PI3D.W = W

        if len(image.shape) == 3:
            nz,nr,nc = image.shape
        elif len(image.shape) == 4: # multi-channel image
            nz,nw,nr,nc = image.shape

        PI3D.NR = nr
        PI3D.NC = nc
        PI3D.NZ = nz

        npr = int(np.ceil(nr/subPatchSize)) # number of patch rows
        npc = int(np.ceil(nc/subPatchSize)) # number of patch cols
        npz = int(np.ceil(nz/subPatchSize)) # number of patch planes

        nrpi = npr*subPatchSize+2*margin # number of rows in padded image 
        ncpi = npc*subPatchSize+2*margin # number of cols in padded image 
        nzpi = npz*subPatchSize+2*margin # number of plns in padded image 

        PI3D.NRPI = nrpi
        PI3D.NCPI = ncpi
        PI3D.NZPI = nzpi

        if len(image.shape) == 3:
            PI3D.PaddedImage = np.zeros((nzpi,nrpi,ncpi))
            PI3D.PaddedImage[margin:margin+nz,margin:margin+nr,margin:margin+nc] = image
        elif len(image.shape) == 4:
            PI3D.PaddedImage = np.zeros((nzpi,nw,nrpi,ncpi))
            PI3D.PaddedImage[margin:margin+nz,:,margin:margin+nr,margin:margin+nc] = image

        PI3D.PC = [] # patch coordinates [z0,z1,r0,r1,c0,c1]
        for iZ in range(npz):
            z0 = iZ*subPatchSize
            z1 = z0+patchSize
            for i in range(npr):
                r0 = i*subPatchSize
                r1 = r0+patchSize
                for j in range(npc):
                    c0 = j*subPatchSize
                    c1 = c0+patchSize
                    PI3D.PC.append([z0,z1,r0,r1,c0,c1])

        PI3D.NumPatches = len(PI3D.PC)
        PI3D.Mode = mode # 'replace' or 'accumulate'

    def getPatch(i):
        z0,z1,r0,r1,c0,c1 = PI3D.PC[i]
        if len(PI3D.PaddedImage.shape) == 3:
            return PI3D.PaddedImage[z0:z1,r0:r1,c0:c1]
        if len(PI3D.PaddedImage.shape) == 4:
            return PI3D.PaddedImage[z0:z1,:,r0:r1,c0:c1]

    def createOutput(nChannels):
        if nChannels == 1:
            PI3D.Output = np.zeros((PI3D.NZPI,PI3D.NRPI,PI3D.NCPI))
        else:
            PI3D.Output = np.zeros((PI3D.NZPI,nChannels,PI3D.NRPI,PI3D.NCPI))
        if PI3D.Mode == 'accumulate':
            PI3D.Count = np.zeros((PI3D.NZPI,PI3D.NRPI,PI3D.NCPI))

    def patchOutput(i,P):
        z0,z1,r0,r1,c0,c1 = PI3D.PC[i]
        if PI3D.Mode == 'accumulate':
            PI3D.Count[z0:z1,r0:r1,c0:c1] += PI3D.W
        if len(P.shape) == 3:
            if PI3D.Mode == 'accumulate':
                PI3D.Output[z0:z1,r0:r1,c0:c1] += np.multiply(P,PI3D.W)
            elif PI3D.Mode == 'replace':
                PI3D.Output[z0:z1,r0:r1,c0:c1] = P
        elif len(P.shape) == 4:
            if PI3D.Mode == 'accumulate':
                for i in range(P.shape[1]):
                    PI3D.Output[z0:z1,i,r0:r1,c0:c1] += np.multiply(P[:,i,:,:],PI3D.W)
            elif PI3D.Mode == 'replace':
                PI3D.Output[z0:z1,:,r0:r1,c0:c1] = P

    def getValidOutput():
        margin = PI3D.Margin
        nz, nr, nc = PI3D.NZ, PI3D.NR, PI3D.NC
        if PI3D.Mode == 'accumulate':
            C = PI3D.Count[margin:margin+nz,margin:margin+nr,margin:margin+nc]
        if len(PI3D.Output.shape) == 3:
            if PI3D.Mode == 'accumulate':
                return np.divide(PI3D.Output[margin:margin+nz,margin:margin+nr,margin:margin+nc],C)
            if PI3D.Mode == 'replace':
                return PI3D.Output[margin:margin+nz,margin:margin+nr,margin:margin+nc]
        if len(PI3D.Output.shape) == 4:
            if PI3D.Mode == 'accumulate':
                for i in range(PI3D.Output.shape[1]):
                    PI3D.Output[margin:margin+nz,i,margin:margin+nr,margin:margin+nc] = np.divide(PI3D.Output[margin:margin+nz,i,margin:margin+nr,margin:margin+nc],C)
            return PI3D.Output[margin:margin+nz,:,margin:margin+nr,margin:margin+nc]


    def demo():
        I = np.random.rand(128,128,128)
        PI3D.setup(I,64,4,'accumulate')

        nChannels = 2
        PI3D.createOutput(nChannels)

        for i in range(PI3D.NumPatches):
            P = PI3D.getPatch(i)
            Q = np.zeros((P.shape[0],nChannels,P.shape[1],P.shape[2]))
            for j in range(nChannels):
                Q[:,j,:,:] = P
            PI3D.patchOutput(i,Q)

        J = PI3D.getValidOutput()
        J = J[:,0,:,:]

        D = np.abs(I-J)
        print(np.max(D))

        pI = I[64,:,:]
        pJ = J[64,:,:]
        pD = D[64,:,:]

        K = cat(1,cat(1,pI,pJ),pD)
        imshow(K)




#import matplotlib.pyplot as plt
import tifffile
import os
import numpy as np
from skimage import io as skio
from scipy.ndimage import *
from skimage.morphology import *
from skimage.transform import resize

def tifread(path):
    return tifffile.imread(path)

def tifwrite(I,path):
    tifffile.imsave(path, I)

def imshow(I,**kwargs):
    if not kwargs:
        plt.imshow(I,cmap='gray')
    else:
        plt.imshow(I,**kwargs)
        
    plt.axis('off')
    plt.show()

def imshowlist(L,**kwargs):
    n = len(L)
    for i in range(n):
        plt.subplot(1, n, i+1)
        if not kwargs:
            plt.imshow(L[i],cmap='gray')
        else:
            plt.imshow(L[i],**kwargs)
        plt.axis('off')
    plt.show()

def imread(path):
    return skio.imread(path)

def imwrite(I,path):
    skio.imsave(path,I)

def im2double(I):
    if I.dtype == 'uint16':
        return I.astype('float64')/65535
    elif I.dtype == 'uint8':
        return I.astype('float64')/255
    elif I.dtype == 'float32':
        return I.astype('float64')
    elif I.dtype == 'float64':
        return I
    else:
        print('returned original image type: ', I.dtype)
        return I

def size(I):
    return list(I.shape)

def imresizeDouble(I,sizeOut): # input and output are double
    return resize(I,(sizeOut[0],sizeOut[1]),mode='reflect')

def imresize3Double(I,sizeOut): # input and output are double
    return resize(I,(sizeOut[0],sizeOut[1],sizeOut[2]),mode='reflect')

def imresizeUInt8(I,sizeOut): # input and output are UInt8
    return np.uint8(resize(I.astype(float),(sizeOut[0],sizeOut[1]),mode='reflect',order=0))

def imresize3UInt8(I,sizeOut): # input and output are UInt8
    return np.uint8(resize(I.astype(float),(sizeOut[0],sizeOut[1],sizeOut[2]),mode='reflect',order=0))

def normalize(I):
    m = np.min(I)
    M = np.max(I)
    if M > m:
        return (I-m)/(M-m)
    else:
        return I

def snormalize(I):
    m = np.mean(I)
    s = np.std(I)
    if s > 0:
        return (I-m)/s
    else:
        return I

def cat(a,I,J):
    return np.concatenate((I,J),axis=a)

def imerode(I,r):
    return binary_erosion(I, disk(r))

def imdilate(I,r):
    return binary_dilation(I, disk(r))

def imerode3(I,r):
    return morphology.binary_erosion(I, ball(r))

def imdilate3(I,r):
    return morphology.binary_dilation(I, ball(r))

def sphericalStructuralElement(imShape,fRadius):
    if len(imShape) == 2:
        return disk(fRadius,dtype=float)
    if len(imShape) == 3:
        return ball(fRadius,dtype=float)

def medfilt(I,filterRadius):
    return median_filter(I,footprint=sphericalStructuralElement(I.shape,filterRadius))

def maxfilt(I,filterRadius):
    return maximum_filter(I,footprint=sphericalStructuralElement(I.shape,filterRadius))

def minfilt(I,filterRadius):
    return minimum_filter(I,footprint=sphericalStructuralElement(I.shape,filterRadius))

def ptlfilt(I,percentile,filterRadius):
    return percentile_filter(I,percentile,footprint=sphericalStructuralElement(I.shape,filterRadius))

def imgaussfilt(I,sigma,**kwargs):
    return gaussian_filter(I,sigma,**kwargs)

def imlogfilt(I,sigma,**kwargs):
    return -gaussian_laplace(I,sigma,**kwargs)

def imgradmag(I,sigma):
    if len(I.shape) == 2:
        dx = imgaussfilt(I,sigma,order=[0,1])
        dy = imgaussfilt(I,sigma,order=[1,0])
        return np.sqrt(dx**2+dy**2)
    if len(I.shape) == 3:
        dx = imgaussfilt(I,sigma,order=[0,0,1])
        dy = imgaussfilt(I,sigma,order=[0,1,0])
        dz = imgaussfilt(I,sigma,order=[1,0,0])
        return np.sqrt(dx**2+dy**2+dz**2)

def localstats(I,radius,justfeatnames=False):
    ptls = [10,30,50,70,90]
    featNames = []
    for i in range(len(ptls)):
        featNames.append('locPtl%d' % ptls[i])
    if justfeatnames == True:
        return featNames
    sI = size(I)
    nFeats = len(ptls)
    F = np.zeros((sI[0],sI[1],nFeats))
    for i in range(nFeats):
        F[:,:,i] = ptlfilt(I,ptls[i],radius)
    return F

def localstats3(I,radius,justfeatnames=False):
    ptls = [10,30,50,70,90]
    featNames = []
    for i in range(len(ptls)):
        featNames.append('locPtl%d' % ptls[i])
    if justfeatnames == True:
        return featNames
    sI = size(I)
    nFeats = len(ptls)
    F = np.zeros((sI[0],sI[1],sI[2],nFeats))
    for i in range(nFeats):
        F[:,:,:,i] = ptlfilt(I,ptls[i],radius)
    return F

def imderivatives(I,sigmas,justfeatnames=False):
    if type(sigmas) is not list:
        sigmas = [sigmas]
    derivPerSigmaFeatNames = ['d0','dx','dy','dxx','dxy','dyy','normGrad','normHessDiag']
    if justfeatnames == True:
        featNames = [];
        for i in range(len(sigmas)):
            for j in range(len(derivPerSigmaFeatNames)):
                featNames.append('derivSigma%d%s' % (sigmas[i],derivPerSigmaFeatNames[j]))
        return featNames
    nDerivativesPerSigma = len(derivPerSigmaFeatNames)
    nDerivatives = len(sigmas)*nDerivativesPerSigma
    sI = size(I)
    D = np.zeros((sI[0],sI[1],nDerivatives))
    for i in range(len(sigmas)):
        sigma = sigmas[i]
        dx = imgaussfilt(I,sigma,order=[0,1])
        dy = imgaussfilt(I,sigma,order=[1,0])
        dxx = imgaussfilt(I,sigma,order=[0,2])
        dyy = imgaussfilt(I,sigma,order=[2,0])
        D[:,:,nDerivativesPerSigma*i  ] = imgaussfilt(I,sigma)
        D[:,:,nDerivativesPerSigma*i+1] = dx
        D[:,:,nDerivativesPerSigma*i+2] = dy
        D[:,:,nDerivativesPerSigma*i+3] = dxx
        D[:,:,nDerivativesPerSigma*i+4] = imgaussfilt(I,sigma,order=[1,1])
        D[:,:,nDerivativesPerSigma*i+5] = dyy
        D[:,:,nDerivativesPerSigma*i+6] = np.sqrt(dx**2+dy**2)
        D[:,:,nDerivativesPerSigma*i+7] = np.sqrt(dxx**2+dyy**2)
    return D
    # derivatives are indexed by the last dimension, which is good for ML features but not for visualization,
    # in which case the expected dimensions are [plane,channel,y(row),x(col)]; to obtain that ordering, do
    # D = np.moveaxis(D,[0,3,1,2],[0,1,2,3])

def imderivatives3(I,sigmas,justfeatnames=False):
    if type(sigmas) is not list:
        sigmas = [sigmas]

    derivPerSigmaFeatNames = ['d0','dx','dy','dz','dxx','dxy','dxz','dyy','dyz','dzz','normGrad','normHessDiag']

    # derivPerSigmaFeatNames = ['d0','normGrad','normHessDiag']

    if justfeatnames == True:
        featNames = [];
        for i in range(len(sigmas)):
            for j in range(len(derivPerSigmaFeatNames)):
                featNames.append('derivSigma%d%s' % (sigmas[i],derivPerSigmaFeatNames[j]))
        return featNames
    nDerivativesPerSigma = len(derivPerSigmaFeatNames)
    nDerivatives = len(sigmas)*nDerivativesPerSigma
    sI = size(I)
    D = np.zeros((sI[0],sI[1],sI[2],nDerivatives)) # plane, channel, y, x
    for i in range(len(sigmas)):
        sigma = sigmas[i]
        dx  = imgaussfilt(I,sigma,order=[0,0,1]) # z, y, x
        dy  = imgaussfilt(I,sigma,order=[0,1,0])
        dz  = imgaussfilt(I,sigma,order=[1,0,0])
        dxx = imgaussfilt(I,sigma,order=[0,0,2])
        dyy = imgaussfilt(I,sigma,order=[0,2,0])
        dzz = imgaussfilt(I,sigma,order=[2,0,0])

        D[:,:,:,nDerivativesPerSigma*i   ] = imgaussfilt(I,sigma)
        D[:,:,:,nDerivativesPerSigma*i+1 ] = dx
        D[:,:,:,nDerivativesPerSigma*i+2 ] = dy
        D[:,:,:,nDerivativesPerSigma*i+3 ] = dz
        D[:,:,:,nDerivativesPerSigma*i+4 ] = dxx
        D[:,:,:,nDerivativesPerSigma*i+5 ] = imgaussfilt(I,sigma,order=[0,1,1])
        D[:,:,:,nDerivativesPerSigma*i+6 ] = imgaussfilt(I,sigma,order=[1,0,1])
        D[:,:,:,nDerivativesPerSigma*i+7 ] = dyy
        D[:,:,:,nDerivativesPerSigma*i+8 ] = imgaussfilt(I,sigma,order=[1,1,0])
        D[:,:,:,nDerivativesPerSigma*i+9 ] = dzz
        D[:,:,:,nDerivativesPerSigma*i+10] = np.sqrt(dx**2+dy**2+dz**2)
        D[:,:,:,nDerivativesPerSigma*i+11] = np.sqrt(dxx**2+dyy**2+dzz**2)

        # D[:,:,:,nDerivativesPerSigma*i   ] = imgaussfilt(I,sigma)
        # D[:,:,:,nDerivativesPerSigma*i+1 ] = np.sqrt(dx**2+dy**2+dz**2)
        # D[:,:,:,nDerivativesPerSigma*i+2 ] = np.sqrt(dxx**2+dyy**2+dzz**2)
    return D
    # derivatives are indexed by the last dimension, which is good for ML features but not for visualization,
    # in which case the expected dimensions are [plane,y(row),x(col)]; to obtain that ordering, do
    # D = np.moveaxis(D,[2,0,1],[0,1,2])

def imfeatures(I=[],sigmaDeriv=1,sigmaLoG=1,locStatsRad=0,justfeatnames=False):
    if type(sigmaDeriv) is not list:
        sigmaDeriv = [sigmaDeriv]
    if type(sigmaLoG) is not list:
        sigmaLoG = [sigmaLoG]
    derivFeatNames = imderivatives([],sigmaDeriv,justfeatnames=True)
    nLoGFeats = len(sigmaLoG)
    locStatsFeatNames = []
    if locStatsRad > 1:
        locStatsFeatNames = localstats([],locStatsRad,justfeatnames=True)
    nLocStatsFeats = len(locStatsFeatNames)
    if justfeatnames == True:
        featNames = derivFeatNames
        for i in range(nLoGFeats):
            featNames.append('logSigma%d' % sigmaLoG[i])
        for i in range(nLocStatsFeats):
            featNames.append(locStatsFeatNames[i])
        return featNames
    nDerivFeats = len(derivFeatNames)
    nFeatures = nDerivFeats+nLoGFeats+nLocStatsFeats
    sI = size(I)
    F = np.zeros((sI[0],sI[1],nFeatures))
    F[:,:,:nDerivFeats] = imderivatives(I,sigmaDeriv)
    for i in range(nLoGFeats):
        F[:,:,nDerivFeats+i] = imlogfilt(I,sigmaLoG[i])
    if locStatsRad > 1:
        F[:,:,nDerivFeats+nLoGFeats:] = localstats(I,locStatsRad)
    return F

def imfeatures3(I=[],sigmaDeriv=2,sigmaLoG=2,locStatsRad=0,justfeatnames=False):
    if type(sigmaDeriv) is not list:
        sigmaDeriv = [sigmaDeriv]
    if type(sigmaLoG) is not list:
        sigmaLoG = [sigmaLoG]
    derivFeatNames = imderivatives3([],sigmaDeriv,justfeatnames=True)
    nLoGFeats = len(sigmaLoG)
    locStatsFeatNames = []
    if locStatsRad > 1:
        locStatsFeatNames = localstats3([],locStatsRad,justfeatnames=True)
    nLocStatsFeats = len(locStatsFeatNames)
    if justfeatnames == True:
        featNames = derivFeatNames
        for i in range(nLoGFeats):
            featNames.append('logSigma%d' % sigmaLoG[i])
        for i in range(nLocStatsFeats):
            featNames.append(locStatsFeatNames[i])
        return featNames
    nDerivFeats = len(derivFeatNames)
    nFeatures = nDerivFeats+nLoGFeats+nLocStatsFeats
    sI = size(I)
    F = np.zeros((sI[0],sI[1],sI[2],nFeatures))
    F[:,:,:,:nDerivFeats] = imderivatives3(I,sigmaDeriv)
    for i in range(nLoGFeats):
        F[:,:,:,nDerivFeats+i] = imlogfilt(I,sigmaLoG[i])
    if locStatsRad > 1:
        F[:,:,:,nDerivFeats+nLoGFeats:] = localstats3(I,locStatsRad)
    return F

def stack2list(S):
    L = []
    for i in range(size(S)[2]):
        L.append(S[:,:,i])
    return L

def thrsegment(I,wsBlr,wsThr): # basic threshold segmentation
    G = imgaussfilt(I,sigma=(1-wsBlr)+wsBlr*5) # min 1, max 5
    M = G > wsThr
    return M


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#Created on Tue Aug 30 08:28:48 2022
#@author: Ajit Johnson Nirmal
#Generation of Thumbnails/ training data for a given marker

"""
!!! abstract "Short Description"
    The `generateThumbnails` function generates Thumbnails of positive and 
    negative cells for a specified marker. The Thumbnails will be used to train a deep learning model. Make sure to have 
    the raw image, computed single-cell spatial table, and markers.csv file 
    ready for input.


## Function
"""


# import packages
import numpy as np
import pandas as pd
import random
import tifffile
import os
import pathlib
import dask.array as da
import argparse
from sklearn.mixture import GaussianMixture
from collections import defaultdict
from sklearn.neighbors import BallTree
#import zarr

# Function
def generateThumbnails (spatialTablePath, 
                        imagePath, 
                        markerChannelMapPath,
                        markers, 
                        markerColumnName='marker',
                        channelColumnName='channel',
                        transformation=True, 
                        maxThumbnails=2000, 
                        random_state=0,
                        localNorm=True, 
                        globalNorm=False,
                        x_coordinate='X_centroid', 
                        y_coordinate='Y_centroid',
                        percentiles=[2, 12, 88, 98], 
                        windowSize=64,
                        restrictDensity=True,
                        restrictDensityNumber=None,
                        verbose=True,
                        projectDir=None):
    """
Parameters:
    spatialTablePath (str):
        Path to the single-cell spatial feature matrix.

    imagePath (str):
        Path to the image file. Recognizes `.ome.tif` image file.
    
    markerChannelMapPath (str):
        Path to a `markers.csv` file that maps the channel number with the marker information. 
        Create a .csv file with at least two columns named 'channel' and 'marker' that 
        map the channel numbers to their corresponding markers. The channel number 
        should use 1-based indexing.
    
    markers (list):
        Markers for which `Thumbnails` need to be generated. The function looks for
        these listed names in the `single-cell spatial Table`.

    markerColumnName (str):
        The name of the column in the `markers.csv` file that holds the marker information. 
    
    channelColumnName (str):
        The name of the column in the `markers.csv` file that holds the channel information.  

    transformation (bool, optional):
        Performs `arcsinh` transformation on the data. If the `single-cell spatial table`
        is already transformed (like log transformation), set this to `False`.

    maxThumbnails (int, optional):
        Maximum number of Thumbnails to generate. 

    random_state (int, optional):
        Seed used by the random number generator.

    localNorm (bool, optional):
        It creates a duplicate folder of the Thumbnails, with local normalization
        performed on the images. Local normalization is the process of dividing
        each pixel in a thumbnail by the maximum value across the entire thumbnail.
        This is helpful for visual supervised sorting of the Thumbnails.

    globalNorm (bool, optional):
        It creates a duplicate folder of the Thumbnails, with global normalization
        performed on the images. Global normalization is the process of dividing
        each pixel in a thumbnail by the maximum value of the given marker across
        the entire image.

    x_coordinate (str, optional):
        The column name in `single-cell spatial table` that records the
        X coordinates for each cell.

    y_coordinate (str, optional):
        The column name in `single-cell spatial table` that records the
        Y coordinates for each cell.

    percentiles (list, optional):
        Specify the interval of percentile levels of the expression utilized to intialize
        the GMM. The cells falling within these percentiles are utilized to distinguish
        between negative cells (first two values) and positive cells (last two values).

    windowSize (int, optional):
        Size of the Thumbnails.

    restrictDensity (bool, optional):
        This parameter is utilized to regulate the number of positive cells 
        observed in a given field of view. In the case of markers that do not 
        exhibit a distinct spatial pattern, such as immune cells, it is 
        recommended to train the model using sparse cells in the field of view.

    restrictDensityNumber (int, optional):
        This parameter is employed in conjunction with `restrictDensity`. 
        By default, the program attempts to automatically identify less dense 
        regions when restrictDensity is set to `True` using a GMM approach. 
        However, `restrictDensityNumber` can be utilized to exert greater 
        control over the process, allowing the user to limit the number of 
        positive cells they wish to observe within the field of view. 
        This parameter requires integers.

    verbose (bool, optional):
        If True, print detailed information about the process to the console.  
        
    projectDir (string, optional):
        Path to output directory. The result will be located at
        `projectDir/CSPOT/Thumbnails/`.

Returns:
    Thumbnails (image):
        Saves Thumbnails of auto identified postive and negative cells the
        designated output directory.

Example:
        ```python
        
        # set the working directory & set paths to the example data
        projectDir = '/Users/aj/Documents/cspotExampleData'
        imagePath = projectDir + '/image/exampleImage.tif'
        spatialTablePath = projectDir + '/quantification/exampleSpatialTable.csv'
        markerChannelMapPath = projectDir + '/markers.csv'
        
        # Run the function
        cs.generateThumbnails ( spatialTablePath=spatialTablePath, 
                        imagePath=imagePath, 
                        markerChannelMapPath=markerChannelMapPath,
                        markers=["ECAD", "CD3D"], 
                        markerColumnName='marker',
                        channelColumnName='channel',
                        transformation=True, 
                        maxThumbnails=100, 
                        random_state=0,
                        localNorm=True, 
                        globalNorm=False,
                        x_coordinate='X_centroid', 
                        y_coordinate='Y_centroid',
                        percentiles=[2, 12, 88, 98], 
                        windowSize=64,
                        restrictDensity=True,
                        restrictDensityNumber=None,
                        verbose=True,
                        projectDir=cwd)
        
        # Same function if the user wants to run it via Command Line Interface
        python generateThumbnails.py \
            --spatialTablePath /Users/aj/Desktop/cspotExampleData/quantification/exampleSpatialTable.csv \
            --imagePath /Users/aj/Desktop/cspotExampleData/image/exampleImage.tif \
            --markerChannelMapPath /Users/aj/Desktop/cspotExampleData/markers.csv \
            --markers ECAD CD3D \
            --maxThumbnails 100 \
            --projectDir /Users/aj/Desktop/cspotExampleData/
        
        ```
    """
    
    # read the markers.csv
    maper = pd.read_csv(pathlib.Path(markerChannelMapPath))
    columnnames =  [word.lower() for word in maper.columns]
    maper.columns = columnnames
    
    # identify the marker column name (doing this to make it easier for people who confuse between marker and markers)
    if markerColumnName not in columnnames:
        if markerColumnName != 'marker':
            raise ValueError('markerColumnName not found in markerChannelMap, please check')
        if 'markers' in columnnames:
            markerCol = 'markers'
        else:
            raise ValueError('markerColumnName not found in markerChannelMap, please check')
    else: 
        markerCol = markerColumnName
    
    # identify the channel column name (doing this to make it easier for people who confuse between channel and channels)
    if channelColumnName not in columnnames:
        if channelColumnName != 'channel':
            raise ValueError('channelColumnName not found in markerChannelMap, please check')
        if 'channels' in columnnames:
            channelCol = 'channels'
        else:
            raise ValueError('channelColumnName not found in markerChannelMap, please check')
    else: 
        channelCol = channelColumnName
    
    # map the marker and channels
    chmamap = dict(zip(maper[markerCol], maper[channelCol]))
    
    # load the CSV to identify potential thumbnails
    data = pd.read_csv(pathlib.Path(spatialTablePath))
    #data.index = data.index.astype(str)

    # subset the markers of interest
    if isinstance (markers, str):
        markers = [markers]
    
    # find the corresponding channel names
    markerChannels = [chmamap[key] for key in markers if key in chmamap]
    # convert markerChannels to zero indexing
    markerChannels = [x-1 for x in markerChannels]

    # creat a dict of marker and corresponding marker channel
    marker_map = dict(zip(markers,markerChannels))

    # create folders if it does not exist
    if projectDir is None:
        projectDir = os.getcwd()

    # TruePos folders
    for i in markers:
        pos_path = pathlib.Path(projectDir + '/CSPOT/Thumbnails/' + str(i) + '/TruePos')
        neg_path = pathlib.Path(projectDir + '/CSPOT/Thumbnails/' + str(i) + '/TrueNeg')
        pos2neg_path = pathlib.Path(projectDir + '/CSPOT/Thumbnails/' + str(i) + '/PosToNeg')
        neg2pos_path = pathlib.Path(projectDir + '/CSPOT/Thumbnails/' + str(i) + '/NegToPos')
        if not os.path.exists(pos_path):
            os.makedirs(pos_path)
        if not os.path.exists(neg_path):
            os.makedirs(neg_path)
        if not os.path.exists(pos2neg_path):
            os.makedirs(pos2neg_path)
        if not os.path.exists(neg2pos_path):
            os.makedirs(neg2pos_path)
        if localNorm is True:
            local_pos_path = pathlib.Path(projectDir + '/CSPOT/Thumbnails/localNorm/' + str(i) + '/TruePos')
            local_neg_path = pathlib.Path(projectDir + '/CSPOT/Thumbnails/localNorm/' + str(i) + '/TrueNeg')
            local_pos2neg_path = pathlib.Path(projectDir + '/CSPOT/Thumbnails/localNorm/' + str(i) + '/PosToNeg')
            local_neg2pos_path = pathlib.Path(projectDir + '/CSPOT/Thumbnails/localNorm/' + str(i) + '/NegToPos')
            if not os.path.exists(local_pos_path):
                os.makedirs(local_pos_path)
            if not os.path.exists(local_neg_path):
                os.makedirs(local_neg_path)
            if not os.path.exists(local_pos2neg_path):
                os.makedirs(local_pos2neg_path)
            if not os.path.exists(local_neg2pos_path):
                os.makedirs(local_neg2pos_path)

    marker_data = data[markers]
    location = data[[x_coordinate,y_coordinate]]

    # clip the data to drop outliers
    def clipping (x):
        clip = x.clip(lower =np.percentile(x,0.01), upper=np.percentile(x,99.99)).tolist()
        return clip
    
    # return the mean between two percentiles
    def meanPercentile (values, lowPercentile, highPercentile):
        # Calculate the 1st percentile value
        p1 = np.percentile(values, lowPercentile)
        # Calculate the 20th percentile value
        p20 = np.percentile(values, highPercentile)
        # Select the values between the 1st and 20th percentile using numpy.where()
        filtered_values = np.where((values >= p1) & (values <= p20))
        # Calculate the mean of the filtered values
        meanVal = np.mean(values[filtered_values])
        return meanVal

    # Function for GMM
    def simpleGMM (data, n_components, means_init, random_state):
        gmm = GaussianMixture(n_components=n_components, means_init=means_init,  random_state=random_state)
        gmm.fit(data)
        # Predict the class labels for each sample
        predictions = gmm.predict(data)
        # Get the mean of each Gaussian component
        means = gmm.means_.flatten()
        # Sort the mean values in ascending order
        sorted_means = np.sort(means)
        # Assign 'pos' to rows with higher mean distribution and 'neg' to rows with lower mean distribution
        labels = np.where(predictions == np.argmax(means), 'pos', 'neg')
        return labels
        
    # match two arrays and return seperate lists
    def array_match (labels, names):
        # Create a defaultdict with an empty list as the default value
        result = defaultdict(list)
        # Iterate over the labels and names arrays
        for label, name in zip(labels, names):
            # Add the name to the list for the corresponding label
            result[label].append(name)
        return result
        
        
    
    # clip data
    marker_data = marker_data.apply(clipping)

    # apply transformation if requested
    if transformation is True:
        marker_data = np.arcsinh(marker_data)
        #marker_data = np.log1p(marker_data)
        
    # combine data
    combined_data = pd.concat([marker_data, location], axis=1)

    # intialize the percentiles values
    percentiles.sort()

    # function to identify the corner of the thumbnails
    def cornerFinder (centroid):
        row_start = int(centroid - windowSize // 2)
        row_end = row_start + windowSize
        return [row_start, row_end]

    # function to crop the image and save the image
    def cropImage (rowIndex, corners, imgType, zimg, npercentile, m, maxpercentile, imname):
        #print(str(rowIndex))
        x_start = corners.loc[rowIndex]['x_start']; x_end = corners.loc[rowIndex]['x_end']
        y_start = corners.loc[rowIndex]['y_start']; y_end = corners.loc[rowIndex]['y_end']
        # cropping image
        crop = zimg[y_start:y_end, x_start:x_end]
        # check if image is the right size
        if crop.shape == (windowSize,windowSize):
            # convert the image to unit8
            if globalNorm is True:
                fullN = ((crop/npercentile)*255).clip(0, 255).astype('uint8')
            else:
                fullN = ((crop/maxpercentile)*255).clip(0, 255).astype('uint8')
            # save the cropped image
            if imgType == 'pos':
                path = pathlib.Path(projectDir + '/CSPOT/Thumbnails/' + str(m) + '/TruePos/' + str(rowIndex) + "_" + str(imname) + '.tif')
            elif imgType == 'neg':
                path = pathlib.Path(projectDir + '/CSPOT/Thumbnails/' + str(m) + '/TrueNeg/' + str(rowIndex) + "_" + str(imname) + '.tif')                      
            # write file
            tifffile.imwrite(path,fullN)
            # local normalization if requested
            if localNorm is True:
                localN = ((crop/(np.percentile(crop, 99.99)))*255).clip(0, 255).astype('uint8') #.compute()
                # save image
                if imgType == 'pos':
                    Lpath = pathlib.Path(projectDir + '/CSPOT/Thumbnails/localNorm/' + str(m) + '/TruePos/' + str(rowIndex) + "_" + str(imname) + '.tif')
                elif imgType == 'neg':
                    Lpath = pathlib.Path(projectDir + '/CSPOT/Thumbnails/localNorm/' + str(m) + '/TrueNeg/' + str(rowIndex) + "_" + str(imname) + '.tif')     
                # write file
                tifffile.imwrite(Lpath,localN)

    # identify the cells of interest
    def processMarker (marker):
        if verbose is True:
            print('Processing Marker: ' + str(marker))

        moi = combined_data[marker].values

        # figure out marker index or channel in image
        markerIndex = marker_map[marker]
        
        # mean of cells within defined threshold
        lowerPercent = meanPercentile (values=moi, lowPercentile=percentiles[0], highPercentile=percentiles[1])
        higherPercent = meanPercentile (values=moi, lowPercentile=percentiles[2], highPercentile=percentiles[3])
        # Format mean to pass into next GMM
        Pmean = np.array([[lowerPercent], [higherPercent]])
        
        # perform GMM
        labels = simpleGMM (data=moi.reshape(-1, 1), n_components=2, means_init=Pmean,  random_state=random_state)
        # Match the labels and index names to identify which cells are pos and neg
        expCells = array_match (labels=labels, names=data.index)
        # split it
        pos = expCells.get('pos', []) ; neg = expCells.get('neg', [])
            
            
        # determine the percentiles value for the marker of interest
        #low_a = np.percentile(moi, percentiles[0]); low_b = np.percentile(moi, percentiles[1])
        #high_a = np.percentile(moi, percentiles[2]); high_b = np.percentile(moi, percentiles[3])

        # identify the cells that fall within the determined range
        #neg = np.where(moi.between(low_a, low_b))[0]
        #pos = np.where(moi.between(high_a, high_b))[0]

        # shuffle the cells
        random.Random(random_state).shuffle(neg); random.Random(random_state).shuffle(pos)

        # identify the location of pos and neg cells
        neg_location_i = location.iloc[neg]
        pos_location_i = location.iloc[pos]
    
        # divide the pos cells into two bins based on the number of neighbours
        # assumption is that we will find cells that are dense and sparse
        if restrictDensity is True:
            # identify postive cells that are densly packed if user requests
            kdt = BallTree(pos_location_i[[x_coordinate,y_coordinate]], metric='euclidean') 
            ind = kdt.query_radius(pos_location_i[[x_coordinate,y_coordinate]], r=windowSize+5, return_distance=False)
            for i in range(0, len(ind)): ind[i] = np.delete(ind[i], np.argwhere(ind[i] == i))#remove self
            neigh_length = [len(subarray) for subarray in ind]
            if restrictDensityNumber is None:
                # GMM for auto detection
                X = np.array(neigh_length).reshape(-1, 1)
                gmm_neigh = GaussianMixture(n_components=2)
                gmm_neigh.fit(X)
                means = gmm_neigh.means_
                index = np.argmin(means)
                labels_neigh = gmm_neigh.predict(X)
                lower_mean_indices = np.where(labels_neigh == index)[0]
                # subset the postive cells based on the index
                pos_location_i = pos_location_i.iloc[lower_mean_indices]
            else:
                lower_mean_indices = [i for i, x in enumerate(neigh_length) if x < restrictDensityNumber]
                pos_location_i = pos_location_i.iloc[lower_mean_indices]

        # Find corner
        # Negative cells
        r_cornerFinder = lambda x: cornerFinder (centroid=x)
        neg_x = pd.DataFrame(list(map(r_cornerFinder, neg_location_i[x_coordinate].values))) # x direction
        neg_y = pd.DataFrame(list(map(r_cornerFinder, neg_location_i[y_coordinate].values))) # y direction
        neg_x.columns = ["x_start", "x_end"]; neg_y.columns = ["y_start", "y_end"]
        neg_location = pd.concat([neg_x, neg_y], axis=1)
        neg_location.index = neg_location_i.index

        # Positive cells
        r_cornerFinder = lambda x: cornerFinder (centroid=x)
        pos_x = pd.DataFrame(list(map(r_cornerFinder, pos_location_i[x_coordinate].values))) # x direction
        pos_y = pd.DataFrame(list(map(r_cornerFinder, pos_location_i[y_coordinate].values))) # y direction
        pos_x.columns = ["x_start", "x_end"]; pos_y.columns = ["y_start", "y_end"]
        pos_location = pd.concat([pos_x, pos_y], axis=1)
        pos_location.index = pos_location_i.index

        # drop all coordinates with neg values (essentially edges of slide)
        neg_location = neg_location[(neg_location > 0).all(1)]
        pos_location = pos_location[(pos_location > 0).all(1)]

        # subset max number of cells
        if len(neg_location) > maxThumbnails:
            neg_location = neg_location[:maxThumbnails]
        if len(pos_location) > maxThumbnails:
            pos_location = pos_location[:maxThumbnails]

        # identify image name
        imname = pathlib.Path(imagePath).stem
        
        # load the image
        zimg = tifffile.imread(str(pathlib.Path(imagePath)), level=0, key=markerIndex)
        npercentile = np.percentile(zimg, 99.99)
        maxpercentile = zimg.max()

        # load the image (dask has issues)
        #zimg = da.from_zarr(tifffile.imread(pathlib.Path(imagePath), aszarr=True, level=0, key=markerIndex))
        #npercentile = np.percentile(zimg.compute(), 99.99)
        #maxpercentile = zimg.max().compute()
        

        # for older version f tifffile
        #zimg = zarr.open(tifffile.imread(pathlib.Path(imagePath), aszarr=True, level=0, key=markerIndex))
        # = np.percentile(zimg,  99.99)

        # Cut images and write it out
        # neg
        r_cropImage = lambda x: cropImage (rowIndex=x, corners=neg_location, imgType='neg', zimg=zimg, npercentile=npercentile, maxpercentile=maxpercentile, m=marker, imname=imname)
        process_neg = list(map(r_cropImage, list(neg_location.index)))
        # pos
        r_cropImage = lambda x: cropImage (rowIndex=x, corners=pos_location, imgType='pos', zimg=zimg, npercentile=npercentile, maxpercentile=maxpercentile, m=marker, imname=imname)
        process_neg = list(map(r_cropImage, list(pos_location.index)))


    # Run the function for each marker
    r_processMarker = lambda x: processMarker (marker=x)
    final = list(map(r_processMarker, markers))

    # Finish Job
    if verbose is True:
        print('Thumbnails have been generated, head over to "' + str(projectDir) + '/CSPOT/Thumbnails" to view results')

# Make the Function CLI compatable
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Generate Thumbnails for single-cell spatial feature matrix and image file")
    parser.add_argument("--spatialTablePath", type=str, help="Path to the single-cell spatial feature matrix.")
    parser.add_argument("--imagePath", type=str, help="Path to the image file. Recognizes `.ome.tif` image file.")
    parser.add_argument("--markerChannelMapPath", type=str, help="Path to the marker channel mapping file.")
    parser.add_argument("--markers", type=str, nargs='+', help="Markers for which `Thumbnails` need to be generated. The function looks for these listed names in the `single-cell spatial feature matrix`.")
    parser.add_argument("--markerColumnName", type=str, default='marker', help="The name of the marker column.")
    parser.add_argument("--channelColumnName", type=str, default='channel', help="The name of the channel column.")
    parser.add_argument("--transformation", type=bool, default=True, help="Performs `arcsinh` transformation on the data. If the `single-cell spatial table` is already transformed (like log transformation), set this to `False`.")
    parser.add_argument("--maxThumbnails", type=int, default=2000, help="Maximum number of Thumbnails to generate.")
    parser.add_argument("--random_state", type=int, default=0, help="Seed used by the random number generator.")
    parser.add_argument("--localNorm", type=bool, default=True, help="It creates a duplicate folder of the Thumbnails, with local normalization performed on the images. Local normalization is the process of dividing each pixel in a thumbnail by the maximum value across the entire thumbnail. This is helpful for visual supervised sorting of the Thumbnails.")
    parser.add_argument("--globalNorm", type=bool, default=False, help="It creates a duplicate folder of the Thumbnails, with global normalization performed on the images. Global normalization is the process of dividing each pixel in a thumbnail by the maximum value of the given marker across the entire image.")
    parser.add_argument("--x_coordinate", type=str, default='X_centroid', help="The column name in `single-cell spatial table` that records the X coordinates for each cell. The default is 'X_centroid'.")
    parser.add_argument("--y_coordinate", type=str, default='Y_centroid', help="The column name in `single-cell spatial table` that records the Y coordinates for each cell. The default is 'Y_centroid'.")
    parser.add_argument("--percentiles", type=int, nargs='+', default=[2, 12, 88, 98], help="Specify the interval of percentile levels of the expression utilized to intialize the GMM. The cells falling within these percentiles are utilized to distinguish between negative cells (first two values) and positive cells (last two values).")
    parser.add_argument("--windowSize", type=int, default=64, help="Size of the Thumbnails.")
    parser.add_argument("--restrictDensity", type=bool, default=True, help="This parameter is utilized to regulate the number of positive cells observed in a given field of view. In the case of markers that do not exhibit a distinct spatial pattern, such as immune cells, it is recommended to train the model using sparse cells in the field of view.")
    parser.add_argument("--restrictDensityNumber", type=int, default=None, help="This parameter is employed in conjunction with `restrictDensity`. By default, the program attempts to automatically identify less dense regions when restrictDensity is set to `True` using a GMM approach. However, `restrictDensityNumber` can be utilized to exert greater control over the process, allowing the user to limit the number of positive cells they wish to observe within the field of view. This parameter requires integers.")
    parser.add_argument("--verbose", type=bool, default=True, help="If True, print detailed information about the process to the console.")    
    parser.add_argument("--projectDir", type=str, default=None, help="Path to output directory. The result will be located at `projectDir/CSPOT/Thumbnails/`.")
    args = parser.parse_args()
    generateThumbnails(spatialTablePath=args.spatialTablePath,
                        imagePath=args.imagePath,
                        markerChannelMapPath=args.markerChannelMapPath,
                        markers=args.markers,
                        markerColumnName=args.markerColumnName,
                        channelColumnName=args.channelColumnName,
                        transformation=args.transformation,
                        maxThumbnails=args.maxThumbnails,
                        random_state=args.random_state,
                        localNorm=args.localNorm,
                        globalNorm=args.globalNorm,
                        x_coordinate=args.x_coordinate,
                        y_coordinate=args.y_coordinate,
                        percentiles=args.percentiles,
                        windowSize=args.windowSize,
                        restrictDensity=args.restrictDensity,
                        restrictDensityNumber=args.restrictDensityNumber,
                        verbose=args.verbose,
                        projectDir=args.projectDir)



from .generateThumbnails import generateThumbnails
from .cloneFolder import cloneFolder
from .generateTrainTestSplit import generateTrainTestSplit
from .generateCSScore import generateCSScore
from .csObject import csObject
from .mergecsObject import mergecsObject
from .cspot import cspot
from .csPhenotype import csPhenotype
from .csTrain import csTrain
from .csPredict import csPredict
from .addPredictions import addPredictions
from .UNet import concat3, UNet2D
from .csPipeline import csPipeline
from .csExport import csExport
from .scatterPlot import scatterPlot